"""A manipulation and optimization library for Scalable Vector Graphics (SVG).

This package provides a set of tools for parsing, manipulating, and serializing
SVG files. It allows you to create, modify, and optimize SVG graphics in a
Pythonic way.

The package has a flat import structure, so you can import any symbol directly
by using the following syntax:
```
from svglab import <symbol>
```

Some stuff to get you started:
- use `parse_svg` to parse an SVG string or file into an `Svg` object,
- use `Svg` directly to create a new SVG document from scratch,
- use `.find()` and `.find_all()` to search for elements in the SVG tree,
- use `.to_xml()` to serialize an SVG element to a string,
- use `.save()` to serialize an SVG document and write it to a file,
- use `Formatter` and `set_formatter()` to customize the serialization of SVG
elements.

Example usage:
```
>>> from svglab import parse_svg, Circle, Rect, Length, Rotate
>>> svg = parse_svg("<svg><circle cx='50' cy='50' r='40'/></svg>")
>>> circle = svg.find(Circle)
>>> circle.cx
Length(value=50.0, unit=None)
>>> _ = svg.add_child(
...    Rect(width=Length(100), height=Length(100), transform=[Rotate(45)])
... )
>>> print(svg.to_xml())
<svg>
  <circle cx="50" cy="50" r="40"/>
  <rect height="100" transform="rotate(45)" width="100"/>
</svg>

```
"""

from importlib import metadata as __metadata

import affine as __affine
from typing_extensions import Final as __Final

from svglab import constants as __constants
from svglab.attrparse.angle import Angle, AngleUnit
from svglab.attrparse.color import Color
from svglab.attrparse.iri import FuncIri, Iri
from svglab.attrparse.length import Length
from svglab.attrparse.path_data import (
    ArcTo,
    ClosePath,
    CubicBezierTo,
    HorizontalLineTo,
    LineTo,
    MoveTo,
    PathCommand,
    PathData,
    QuadraticBezierTo,
    SmoothCubicBezierTo,
    SmoothQuadraticBezierTo,
    VerticalLineTo,
)
from svglab.attrparse.point import Point
from svglab.attrparse.points import Points
from svglab.attrparse.transform import (
    Matrix,
    Reifiable,
    Rotate,
    Scale,
    SkewX,
    SkewY,
    Transform,
    TransformFunction,
    Translate,
    compose,
)
from svglab.attrs.attrdefs import (
    AccentHeightAttr,
    AccumulateAttr,
    AdditiveAttr,
    AlignmentBaselineAttr,
    AlphabeticAttr,
    AmplitudeAttr,
    ArabicFormAttr,
    AscentAttr,
    AttributeNameAttr,
    AttributeTypeAttr,
    AzimuthAttr,
    BaseFrequencyAttr,
    BaselineShiftAttr,
    BaseProfileAttr,
    BboxAttr,
    BeginAttr,
    BiasAttr,
    ByAttr,
    CalcModeAttr,
    CapHeightAttr,
    ClassAttr,
    ClipAttr,
    ClipPathAttr,
    ClipPathUnitsAttr,
    ClipRuleAttr,
    ColorAttr,
    ColorInterpolationAttr,
    ColorInterpolationFiltersAttr,
    ColorProfileAttr,
    ColorRenderingAttr,
    ContentScriptTypeAttr,
    ContentStyleTypeAttr,
    CursorAttr,
    CxAttr,
    CyAttr,
    DAttr,
    DescentAttr,
    DiffuseConstantAttr,
    DirectionAttr,
    DisplayAttr,
    DivisorAttr,
    DominantBaselineAttr,
    DurAttr,
    DxListOfLengthsAttr,
    DxNumberAttr,
    DyListOfLengthsAttr,
    DyNumberAttr,
    EdgeModeAttr,
    ElevationAttr,
    EnableBackgroundAttr,
    EndAttr,
    ExponentAttr,
    ExternalResourcesRequiredAttr,
    FillAttr,
    FillOpacityAttr,
    FillRuleAttr,
    FilterAttr,
    FilterResAttr,
    FilterUnitsAttr,
    FloodColorAttr,
    FloodOpacityAttr,
    FontFamilyAttr,
    FontSizeAdjustAttr,
    FontSizeAttr,
    FontStretchAttr,
    FontStyleAttr,
    FontVariantAttr,
    FontWeightAttr,
    FormatAttr,
    FrAttr,
    FromAttr,
    FxAttr,
    FyAttr,
    G1Attr,
    G2Attr,
    GlyphNameAttr,
    GlyphOrientationHorizontalAttr,
    GlyphOrientationVerticalAttr,
    GlyphRefAttr,
    GradientTransformAttr,
    GradientUnitsAttr,
    HangingAttr,
    HeightAttr,
    HorizAdvXAttr,
    HorizOriginXAttr,
    HorizOriginYAttr,
    HrefAttr,
    IdAttr,
    IdeographicAttr,
    ImageRenderingAttr,
    In2Attr,
    InAttr,
    InterceptAttr,
    K1Attr,
    K2Attr,
    K3Attr,
    K4Attr,
    KAttr,
    KernelMatrixAttr,
    KernelUnitLengthAttr,
    KerningAttr,
    KeyPointsAttr,
    KeySplinesAttr,
    KeyTimesAttr,
    LangAttr,
    LangGlyphAttr,
    LengthAdjustAttr,
    LetterSpacingAttr,
    LightingColorAttr,
    LimitingConeAngleAttr,
    LocalAttr,
    MarkerEndAttr,
    MarkerHeightAttr,
    MarkerMidAttr,
    MarkerStartAttr,
    MarkerUnitsAttr,
    MarkerWidthAttr,
    MaskAttr,
    MaskContentUnitsAttr,
    MaskUnitsAttr,
    MathematicalAttr,
    MaxAttr,
    MediaAttr,
    MethodAttr,
    MinAttr,
    ModeAttr,
    NameAnythingAttr,
    NameNameAttr,
    NumOctavesAttr,
    OffsetNumberAttr,
    OffsetNumberPercentageAttr,
    OnAbortAttr,
    OnActivateAttr,
    OnBeginAttr,
    OnClickAttr,
    OnEndAttr,
    OnErrorAttr,
    OnFocusInAttr,
    OnFocusOutAttr,
    OnLoadAttr,
    OnMouseDownAttr,
    OnMouseMoveAttr,
    OnMouseOutAttr,
    OnMouseOverAttr,
    OnMouseUpAttr,
    OnRepeatAttr,
    OnResizeAttr,
    OnScrollAttr,
    OnUnloadAttr,
    OnZoomAttr,
    OpacityAttr,
    OperatorFeCompositeAttr,
    OperatorFeMorphologyAttr,
    OrderAttr,
    OrientationAttr,
    OrientAttr,
    OriginAttr,
    OverflowAttr,
    OverlinePositionAttr,
    OverlineThicknessAttr,
    PaintOrderAttr,
    Panose1Attr,
    PathAttr,
    PathLengthAttr,
    PatternContentUnitsAttr,
    PatternTransformAttr,
    PatternUnitsAttr,
    PointerEventsAttr,
    PointsAttr,
    PointsAtXAttr,
    PointsAtYAttr,
    PointsAtZAttr,
    PreserveAlphaAttr,
    PreserveAspectRatioAttr,
    PrimitiveUnitsAttr,
    RadiusAttr,
    RAttr,
    RefXAttr,
    RefYAttr,
    RenderingIntentAttr,
    RepeatCountAttr,
    RepeatDurAttr,
    RequiredExtensionsAttr,
    RequiredFeaturesAttr,
    RestartAttr,
    ResultAttr,
    RotateListOfNumbersAttr,
    RotateNumberAutoAutoReverseAttr,
    RxAttr,
    RyAttr,
    ScaleAttr,
    SeedAttr,
    ShapeRenderingAttr,
    SlopeAttr,
    SpacingAttr,
    SpecularConstantAttr,
    SpecularExponentAttr,
    SpreadMethodAttr,
    StartOffsetAttr,
    StdDeviationAttr,
    StemhAttr,
    StemvAttr,
    StitchTilesAttr,
    StopColorAttr,
    StopOpacityAttr,
    StrikethroughPositionAttr,
    StrikethroughThicknessAttr,
    StringAttr,
    StrokeAttr,
    StrokeDasharrayAttr,
    StrokeDashoffsetAttr,
    StrokeLinecapAttr,
    StrokeLinejoinAttr,
    StrokeMiterlimitAttr,
    StrokeOpacityAttr,
    StrokeWidthAttr,
    StyleAttr,
    SurfaceScaleAttr,
    SystemLanguageAttr,
    TableValuesAttr,
    TargetAttr,
    TargetXAttr,
    TargetYAttr,
    TextAlignAllAttr,
    TextAlignAttr,
    TextAlignLastAttr,
    TextAnchorAttr,
    TextDecorationAttr,
    TextIndentAttr,
    TextLengthAttr,
    TextRenderingAttr,
    TitleAttr,
    ToAttr,
    TransformAttr,
    TransformOriginAttr,
    TypeAnimateTransformAttr,
    TypeContentTypeAttr,
    TypeFeColorMatrixAttr,
    TypeFeFuncAttr,
    TypeFeTurbluenceAttr,
    U1Attr,
    U2Attr,
    UnderlinePositionAttr,
    UnderlineThicknessAttr,
    UnicodeAttr,
    UnicodeBidiAttr,
    UnicodeRangeAttr,
    UnitsPerEmAttr,
    VAlphabeticAttr,
    ValuesListAttr,
    ValuesListOfNumbersAttr,
    VectorEffectAttr,
    VersionAttr,
    VertAdvYAttr,
    VertOriginXAttr,
    VertOriginYAttr,
    VHangingAttr,
    VIdeographicAttr,
    ViewBoxAttr,
    ViewTargetAttr,
    VisibilityAttr,
    VMathematicalAttr,
    WhiteSpaceAttr,
    WidthAttr,
    WidthsAttr,
    WordSpacingAttr,
    WritingModeAttr,
    X1Attr,
    X2Attr,
    XChannelSelectorAttr,
    XCoordinateAttr,
    XHeightAttr,
    XlinkActuateOnLoadAttr,
    XlinkActuateOnRequestAttr,
    XlinkArcroleAttr,
    XlinkHrefAttr,
    XlinkRoleAttr,
    XlinkShowAttr,
    XlinkTitleAttr,
    XlinkTypeAttr,
    XListOfCoordinatesAttr,
    XmlBaseAttr,
    XmlLangAttr,
    XmlnsAttr,
    XmlSpaceAttr,
    XNumberAttr,
    Y1Attr,
    Y2Attr,
    YChannelSelectorAttr,
    YCoordinateAttr,
    YListOfCoordinatesAttr,
    YNumberAttr,
    ZAttr,
    ZoomAndPanAttr,
)
from svglab.elements.elements import (
    A,
    AltGlyph,
    AltGlyphDef,
    AltGlyphItem,
    Animate,
    AnimateColor,
    AnimateMotion,
    AnimateTransform,
    Circle,
    ClipPath,
    ColorProfile,
    Cursor,
    Defs,
    Desc,
    Ellipse,
    FeBlend,
    FeColorMatrix,
    FeComponentTransfer,
    FeComposite,
    FeConvolveMatrix,
    FeDiffuseLighting,
    FeDisplacementMap,
    FeDistantLight,
    FeFlood,
    FeFuncA,
    FeFuncB,
    FeFuncG,
    FeFuncR,
    FeGaussianBlur,
    FeImage,
    FeMerge,
    FeMergeNode,
    FeMorphology,
    FeOffset,
    FePointLight,
    FeSpecularLighting,
    FeSpotLight,
    FeTile,
    FeTurbulence,
    Filter,
    Font,
    FontFace,
    FontFaceFormat,
    FontFaceName,
    FontFaceSrc,
    FontFaceUri,
    ForeignObject,
    G,
    Glyph,
    GlyphRef,
    Hkern,
    Image,
    Line,
    LinearGradient,
    Marker,
    Mask,
    Metadata,
    MissingGlyph,
    Mpath,
    Path,
    Pattern,
    Polygon,
    Polyline,
    RadialGradient,
    Rect,
    Script,
    Set,
    Stop,
    Style,
    Svg,
    Switch,
    Symbol,
    Text,
    TextPath,
    Title,
    Tref,
    Tspan,
    Use,
    View,
    Vkern,
)
from svglab.elements.traits import (
    BasicShape,
    ContainerElement,
    DescriptiveElement,
    FilterPrimitiveElement,
    GradientElement,
    GraphicsElement,
    GraphicsReferencingElement,
    LightSourceElement,
    Shape,
    StructuralElement,
    TextContentBlockElement,
    TextContentChildElement,
    TextContentElement,
)
from svglab.entities import (
    CData,
    CharacterData,
    Comment,
    Element,
    Entity,
    RawText,
    UnknownElement,
    swap_transforms,
)
from svglab.errors import (
    SvgElementNotFoundError,
    SvgError,
    SvgPathError,
    SvgPathMissingMoveToError,
    SvgTransformSwapError,
    SvgUnitConversionError,
)
from svglab.parse import parse_svg
from svglab.serialize import (
    DEFAULT_FORMATTER,
    MINIMAL_FORMATTER,
    FloatPrecisionSettings,
    Formatter,
    get_current_formatter,
    set_formatter,
)
from svglab.utiltypes import LengthUnit


__all__: __Final = [
    "DEFAULT_FORMATTER",
    "MINIMAL_FORMATTER",
    "A",
    "AccentHeightAttr",
    "AccumulateAttr",
    "AdditiveAttr",
    "AlignmentBaselineAttr",
    "AlphabeticAttr",
    "AltGlyph",
    "AltGlyphDef",
    "AltGlyphItem",
    "AmplitudeAttr",
    "Angle",
    "AngleUnit",
    "Animate",
    "AnimateColor",
    "AnimateMotion",
    "AnimateTransform",
    "ArabicFormAttr",
    "ArcTo",
    "AscentAttr",
    "AttributeNameAttr",
    "AttributeTypeAttr",
    "AzimuthAttr",
    "BaseFrequencyAttr",
    "BaseProfileAttr",
    "BaselineShiftAttr",
    "BasicShape",
    "BboxAttr",
    "BeginAttr",
    "BiasAttr",
    "ByAttr",
    "CData",
    "CalcModeAttr",
    "CapHeightAttr",
    "CharacterData",
    "Circle",
    "ClassAttr",
    "ClipAttr",
    "ClipPath",
    "ClipPathAttr",
    "ClipPathUnitsAttr",
    "ClipRuleAttr",
    "ClosePath",
    "Color",
    "ColorAttr",
    "ColorInterpolationAttr",
    "ColorInterpolationFiltersAttr",
    "ColorProfile",
    "ColorProfileAttr",
    "ColorRenderingAttr",
    "Comment",
    "ContainerElement",
    "ContentScriptTypeAttr",
    "ContentStyleTypeAttr",
    "CubicBezierTo",
    "Cursor",
    "CursorAttr",
    "CxAttr",
    "CyAttr",
    "DAttr",
    "Defs",
    "Desc",
    "DescentAttr",
    "DescriptiveElement",
    "DiffuseConstantAttr",
    "DirectionAttr",
    "DisplayAttr",
    "DivisorAttr",
    "DominantBaselineAttr",
    "DurAttr",
    "DxListOfLengthsAttr",
    "DxNumberAttr",
    "DyListOfLengthsAttr",
    "DyNumberAttr",
    "EdgeModeAttr",
    "Element",
    "ElevationAttr",
    "Ellipse",
    "EnableBackgroundAttr",
    "EndAttr",
    "Entity",
    "ExponentAttr",
    "ExternalResourcesRequiredAttr",
    "FeBlend",
    "FeColorMatrix",
    "FeComponentTransfer",
    "FeComposite",
    "FeConvolveMatrix",
    "FeDiffuseLighting",
    "FeDisplacementMap",
    "FeDistantLight",
    "FeFlood",
    "FeFuncA",
    "FeFuncB",
    "FeFuncG",
    "FeFuncR",
    "FeGaussianBlur",
    "FeImage",
    "FeMerge",
    "FeMergeNode",
    "FeMorphology",
    "FeOffset",
    "FePointLight",
    "FeSpecularLighting",
    "FeSpotLight",
    "FeTile",
    "FeTurbulence",
    "FillAttr",
    "FillOpacityAttr",
    "FillRuleAttr",
    "Filter",
    "FilterAttr",
    "FilterPrimitiveElement",
    "FilterResAttr",
    "FilterUnitsAttr",
    "FloatPrecisionSettings",
    "FloodColorAttr",
    "FloodOpacityAttr",
    "Font",
    "FontFace",
    "FontFaceFormat",
    "FontFaceName",
    "FontFaceSrc",
    "FontFaceUri",
    "FontFamilyAttr",
    "FontSizeAdjustAttr",
    "FontSizeAttr",
    "FontStretchAttr",
    "FontStyleAttr",
    "FontVariantAttr",
    "FontWeightAttr",
    "ForeignObject",
    "FormatAttr",
    "Formatter",
    "FrAttr",
    "FromAttr",
    "FuncIri",
    "FxAttr",
    "FyAttr",
    "G",
    "G1Attr",
    "G2Attr",
    "Glyph",
    "GlyphNameAttr",
    "GlyphOrientationHorizontalAttr",
    "GlyphOrientationVerticalAttr",
    "GlyphRef",
    "GlyphRefAttr",
    "GradientElement",
    "GradientTransformAttr",
    "GradientUnitsAttr",
    "GraphicsElement",
    "GraphicsReferencingElement",
    "HangingAttr",
    "HeightAttr",
    "Hkern",
    "HorizAdvXAttr",
    "HorizOriginXAttr",
    "HorizOriginYAttr",
    "HorizontalLineTo",
    "HrefAttr",
    "IdAttr",
    "IdeographicAttr",
    "Image",
    "ImageRenderingAttr",
    "In2Attr",
    "InAttr",
    "InterceptAttr",
    "Iri",
    "K1Attr",
    "K2Attr",
    "K3Attr",
    "K4Attr",
    "KAttr",
    "KernelMatrixAttr",
    "KernelUnitLengthAttr",
    "KerningAttr",
    "KeyPointsAttr",
    "KeySplinesAttr",
    "KeyTimesAttr",
    "LangAttr",
    "LangGlyphAttr",
    "Length",
    "LengthAdjustAttr",
    "LengthUnit",
    "LetterSpacingAttr",
    "LightSourceElement",
    "LightingColorAttr",
    "LimitingConeAngleAttr",
    "Line",
    "LineTo",
    "LinearGradient",
    "LocalAttr",
    "Marker",
    "MarkerEndAttr",
    "MarkerHeightAttr",
    "MarkerMidAttr",
    "MarkerStartAttr",
    "MarkerUnitsAttr",
    "MarkerWidthAttr",
    "Mask",
    "MaskAttr",
    "MaskContentUnitsAttr",
    "MaskUnitsAttr",
    "MathematicalAttr",
    "Matrix",
    "MaxAttr",
    "MediaAttr",
    "Metadata",
    "MethodAttr",
    "MinAttr",
    "MissingGlyph",
    "ModeAttr",
    "MoveTo",
    "Mpath",
    "NameAnythingAttr",
    "NameNameAttr",
    "NumOctavesAttr",
    "OffsetNumberAttr",
    "OffsetNumberPercentageAttr",
    "OnAbortAttr",
    "OnActivateAttr",
    "OnBeginAttr",
    "OnClickAttr",
    "OnEndAttr",
    "OnErrorAttr",
    "OnFocusInAttr",
    "OnFocusOutAttr",
    "OnLoadAttr",
    "OnMouseDownAttr",
    "OnMouseMoveAttr",
    "OnMouseOutAttr",
    "OnMouseOverAttr",
    "OnMouseUpAttr",
    "OnRepeatAttr",
    "OnResizeAttr",
    "OnScrollAttr",
    "OnUnloadAttr",
    "OnZoomAttr",
    "OpacityAttr",
    "OperatorFeCompositeAttr",
    "OperatorFeMorphologyAttr",
    "OrderAttr",
    "OrientAttr",
    "OrientationAttr",
    "OriginAttr",
    "OverflowAttr",
    "OverlinePositionAttr",
    "OverlineThicknessAttr",
    "PaintOrderAttr",
    "Panose1Attr",
    "Path",
    "PathAttr",
    "PathCommand",
    "PathData",
    "PathLengthAttr",
    "Pattern",
    "PatternContentUnitsAttr",
    "PatternTransformAttr",
    "PatternUnitsAttr",
    "Point",
    "PointerEventsAttr",
    "Points",
    "PointsAtXAttr",
    "PointsAtYAttr",
    "PointsAtZAttr",
    "PointsAttr",
    "Polygon",
    "Polyline",
    "PreserveAlphaAttr",
    "PreserveAspectRatioAttr",
    "PrimitiveUnitsAttr",
    "QuadraticBezierTo",
    "RAttr",
    "RadialGradient",
    "RadiusAttr",
    "RawText",
    "Rect",
    "RefXAttr",
    "RefYAttr",
    "Reifiable",
    "RenderingIntentAttr",
    "RepeatCountAttr",
    "RepeatDurAttr",
    "RequiredExtensionsAttr",
    "RequiredFeaturesAttr",
    "RestartAttr",
    "ResultAttr",
    "Rotate",
    "RotateListOfNumbersAttr",
    "RotateNumberAutoAutoReverseAttr",
    "RxAttr",
    "RyAttr",
    "Scale",
    "ScaleAttr",
    "Script",
    "SeedAttr",
    "Set",
    "Shape",
    "ShapeRenderingAttr",
    "SkewX",
    "SkewY",
    "SlopeAttr",
    "SmoothCubicBezierTo",
    "SmoothQuadraticBezierTo",
    "SpacingAttr",
    "SpecularConstantAttr",
    "SpecularExponentAttr",
    "SpreadMethodAttr",
    "StartOffsetAttr",
    "StdDeviationAttr",
    "StemhAttr",
    "StemvAttr",
    "StitchTilesAttr",
    "Stop",
    "StopColorAttr",
    "StopOpacityAttr",
    "StrikethroughPositionAttr",
    "StrikethroughThicknessAttr",
    "StringAttr",
    "StrokeAttr",
    "StrokeDasharrayAttr",
    "StrokeDashoffsetAttr",
    "StrokeLinecapAttr",
    "StrokeLinejoinAttr",
    "StrokeMiterlimitAttr",
    "StrokeOpacityAttr",
    "StrokeWidthAttr",
    "StructuralElement",
    "Style",
    "StyleAttr",
    "SurfaceScaleAttr",
    "Svg",
    "SvgElementNotFoundError",
    "SvgError",
    "SvgPathError",
    "SvgPathMissingMoveToError",
    "SvgTransformSwapError",
    "SvgUnitConversionError",
    "Switch",
    "Symbol",
    "SystemLanguageAttr",
    "TableValuesAttr",
    "TargetAttr",
    "TargetXAttr",
    "TargetYAttr",
    "Text",
    "TextAlignAllAttr",
    "TextAlignAttr",
    "TextAlignLastAttr",
    "TextAnchorAttr",
    "TextContentBlockElement",
    "TextContentChildElement",
    "TextContentElement",
    "TextDecorationAttr",
    "TextIndentAttr",
    "TextLengthAttr",
    "TextPath",
    "TextRenderingAttr",
    "Title",
    "TitleAttr",
    "ToAttr",
    "Transform",
    "TransformAttr",
    "TransformFunction",
    "TransformOriginAttr",
    "Translate",
    "Tref",
    "Tspan",
    "TypeAnimateTransformAttr",
    "TypeContentTypeAttr",
    "TypeFeColorMatrixAttr",
    "TypeFeFuncAttr",
    "TypeFeTurbluenceAttr",
    "U1Attr",
    "U2Attr",
    "UnderlinePositionAttr",
    "UnderlineThicknessAttr",
    "UnicodeAttr",
    "UnicodeBidiAttr",
    "UnicodeRangeAttr",
    "UnitsPerEmAttr",
    "UnknownElement",
    "Use",
    "VAlphabeticAttr",
    "VHangingAttr",
    "VIdeographicAttr",
    "VMathematicalAttr",
    "ValuesListAttr",
    "ValuesListOfNumbersAttr",
    "VectorEffectAttr",
    "VersionAttr",
    "VertAdvYAttr",
    "VertOriginXAttr",
    "VertOriginYAttr",
    "VerticalLineTo",
    "View",
    "ViewBoxAttr",
    "ViewTargetAttr",
    "VisibilityAttr",
    "Vkern",
    "WhiteSpaceAttr",
    "WidthAttr",
    "WidthsAttr",
    "WordSpacingAttr",
    "WritingModeAttr",
    "X1Attr",
    "X2Attr",
    "XChannelSelectorAttr",
    "XCoordinateAttr",
    "XHeightAttr",
    "XListOfCoordinatesAttr",
    "XNumberAttr",
    "XlinkActuateOnLoadAttr",
    "XlinkActuateOnRequestAttr",
    "XlinkArcroleAttr",
    "XlinkHrefAttr",
    "XlinkRoleAttr",
    "XlinkShowAttr",
    "XlinkTitleAttr",
    "XlinkTypeAttr",
    "XmlBaseAttr",
    "XmlLangAttr",
    "XmlSpaceAttr",
    "XmlnsAttr",
    "Y1Attr",
    "Y2Attr",
    "YChannelSelectorAttr",
    "YCoordinateAttr",
    "YListOfCoordinatesAttr",
    "YNumberAttr",
    "ZAttr",
    "ZoomAndPanAttr",
    "compose",
    "get_current_formatter",
    "parse_svg",
    "set_formatter",
    "swap_transforms",
]


__affine.set_epsilon(__constants.FLOAT_ABSOLUTE_TOLERANCE)

__version__: __Final = __metadata.version(__name__)



"""Miscellaneous utility functions that don't fit anywhere else."""

import re
from collections.abc import Generator

from typing_extensions import TypeIs, TypeVar


_T = TypeVar("_T")


def get_all_subclasses(cls: type[_T], /) -> Generator[type[_T]]:
    """Recursively obtain all subclasses of a class.

    Args:
        cls: The class to obtain subclasses for.

    Yields:
        Subclasses of the given class, including subclasses of subclasses
        (and so on).

    Examples:
        >>> class A:
        ...     pass
        >>> class B(A):
        ...     pass
        >>> class C(B):
        ...     pass
        >>> list(cls.__name__ for cls in get_all_subclasses(A))
        ['B', 'C']

    """
    for subclass in cls.__subclasses__():
        yield subclass
        yield from get_all_subclasses(subclass)


def basic_compare(other: object, /, *, self: _T) -> TypeIs[_T]:
    """Perform a basic comparison between two objects.

    This is a helper function for implementing the `__eq__` method.

    It checks if the other object is the same as the current object or
    an instance of the same class.

    Args:
        other: The other object to compare.
        self: The current object.

    Returns:
        `True` if the other object is the same as the current object or an
        instance of the same class, `False` otherwise.

    Examples:
        >>> class A:
        ...     def __eq__(self, other: object, /) -> bool:
        ...         return basic_compare(other, self=self)
        >>> A() == A()
        True
        >>> A() == 1
        False

    """
    return other is self or isinstance(other, type(self))


def extract_function_name_and_args(attr: str) -> tuple[str, str] | None:
    """Extract function name and arguments from a function-call-like attribute.

    An attribute is considered to be a function call if it has the form
    `name(args)`. This function extracts the name and the arguments from such
    an attribute. If the attribute is not a function call, `None` is returned.

    Args:
    attr: The attribute to extract the function name and arguments from.

    Returns:
    A tuple containing the function name and the arguments,
    or `None` if the attribute is not a function call.

    Examples:
    >>> extract_function_name_and_args("foo()") is None  # no arguments
    True
    >>> extract_function_name_and_args("foo(42)")
    ('foo', '42')
    >>> extract_function_name_and_args("foo(42, 'bar')")
    ('foo', "42, 'bar'")
    >>> extract_function_name_and_args(
    ...     "bar"
    ... ) is None  # not a function call
    True

    """
    match = re.match(r"^([^\(\)]+)\(([^\(\)]+)\)$", attr)

    return None if match is None else (match[1], match[2])



"""Logic for parsing SVG documents."""

import collections
import warnings

import bs4
from typing_extensions import Final, Literal, TypeAlias, cast

from svglab import entities, protocols
from svglab.elements import elements, names
from svglab.utils import miscutils


warnings.filterwarnings("ignore", category=bs4.XMLParsedAsHTMLWarning)

_Markup: TypeAlias = (
    str
    | bytes
    | protocols.SupportsRead[str]
    | protocols.SupportsRead[bytes]
)


_ELEMENT_NAME_TO_CLASS: Final = {
    entities.element_name(cls()): cls
    for cls in miscutils.get_all_subclasses(entities.Element)
    if cls.__name__ in names.ELEMENT_NAME_TO_NORMALIZED.inverse
}

_BS_TO_TEXT_ELEMENT: Final[
    dict[
        type[bs4.NavigableString],
        type[entities.CData | entities.Comment | entities.RawText],
    ]
] = {
    bs4.CData: entities.CData,
    bs4.Comment: entities.Comment,
    bs4.NavigableString: entities.RawText,
}


def _get_root_svg_fragments(soup: bs4.Tag) -> list[bs4.Tag]:
    """Find all root SVG fragments in the given BeautifulSoup object.

    The function performs a breadth-first search until it finds an
    SVG fragment. It then returns a list of all SVG fragments found in the same
    depth of the tree. This allows us to consider an SVG fragment as a root
    element when using HTML parsers that implicitly wrap the document in
    certain HTML elements (e.g., <html>).

    Args:
        soup: A BeautifulSoup `Tag` object representing the root of the
        document.

    Returns:
        A list of SVG fragments found in the document.

    Examples:
        >>> soup = bs4.BeautifulSoup(
        ...     "<svg><rect/></svg>", features="lxml-xml"
        ... )
        >>> _get_root_svg_fragments(soup)
        [<svg><rect/></svg>]
        >>> soup = bs4.BeautifulSoup(
        ...     "<svg><rect/></svg>", features="html.parser"
        ... )
        >>> _get_root_svg_fragments(soup)
        [<svg><rect></rect></svg>]

    """
    queue: collections.deque[bs4.Tag] = collections.deque([soup])

    while queue:
        node = queue.popleft()

        if svg_fragments := node.find_all("svg"):
            return svg_fragments

        queue.extend(
            child for child in node.children if isinstance(child, bs4.Tag)
        )

    return []


def _convert_element(backend: bs4.PageElement) -> entities.Entity | None:
    """Convert a BeautifulSoup element to an `Element` instance.

    Args:
        backend: A BeautifulSoup element to convert.

    Returns:
        An `Element` instance representing the given BeautifulSoup element.

    Raises:
        TypeError: If the given element cannot be converted.

    """
    match backend:
        case bs4.NavigableString():
            cls = _BS_TO_TEXT_ELEMENT.get(type(backend))

            if cls is None:
                return None

            return None if backend.isspace() else cls(backend)
        case bs4.Tag():
            element_class = _ELEMENT_NAME_TO_CLASS.get(
                cast(names.ElementName, backend.name),
                entities.UnknownElement,
            )

            attrs = {"prefix": backend.prefix} | {
                k: str(v).strip() for k, v in backend.attrs.items()
            }

            if element_class is entities.UnknownElement:
                attrs["element_name"] = backend.name

            element = element_class.model_validate(attrs, strict=False)

            for child in backend.children:
                grandchild = _convert_element(child)

                if grandchild is not None:
                    element.add_child(grandchild)
            return element
        case _:
            return None


def _get_markup_head(markup: _Markup, limit: int = 30) -> str:
    match markup:
        case str():
            return (
                markup if len(markup) <= limit else f"{markup[:limit]}..."
            )
        case bytes():
            return _get_markup_head(markup.decode(), limit=limit)
        case protocols.SupportsRead():
            return _get_markup_head(markup.read(limit + 1), limit=limit)


def parse_svg(
    markup: _Markup,
    /,
    *,
    parser: Literal[
        "html.parser", "lxml", "lxml-xml", "html5lib"
    ] = "lxml-xml",
) -> elements.Svg:
    """Parse an SVG document.

    The document must be a valid XML document containing a single SVG
    document fragment.

    Args:
        markup: A string or a file-like object representing markup
        to be parsed.
        parser: The name of the parser to use. Defaults to 'lxml-xml'.

    Returns:
        The parsed SVG document in the form of an `Svg` instance.

    Raises:
        ValueError: If the markup does not contain a single SVG
        document fragment

    Examples:
        >>> svg = parse_svg("<svg><rect/></svg>")
        >>> type(svg).__name__
        'Svg'

    """
    soup = bs4.BeautifulSoup(markup, features=parser)
    svg_fragments = _get_root_svg_fragments(soup)

    if len(svg_fragments) != 1:
        markup_head = _get_markup_head(markup)
        msg = (
            f"Expected one <svg> element, found {len(svg_fragments)}; this "
            f"does not look like a well-formed SVG (markup: {markup_head})."
        )

        raise ValueError(msg)

    return cast(elements.Svg, _convert_element(svg_fragments[0]))



"""Utilities for working with Pydantic."""

import copy
import functools
import re
import reprlib
from collections.abc import Callable

import pydantic
from typing_extensions import Annotated, TypeAlias, TypeVar, override


_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_ListOrTupleT = TypeVar("_ListOrTupleT", list[str], tuple[str])
_BaseModelT = TypeVar("_BaseModelT", bound=pydantic.BaseModel)

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")


Attr: TypeAlias = _T_co | None
""" Pydantic field for an attribute. """


def convert(
    source: pydantic.BaseModel,
    target_type: type[_BaseModelT],
    *,
    strict: bool | None = None,
    deepcopy: bool = True,
) -> _BaseModelT:
    """Convert a pydantic model to another pydantic model.

    Fields that are defined on both models are copied over. If the source model
    has extra fields, they are included in the new model as well. Other fields
    are discarded.

    If `deepcopy` is set to `False`, the source model instance should not be
    used after the conversion, as the new model will hold references to the
    source model's fields.

    Args:
        source: The source model to convert.
        target_type: The target model type.
        strict: Whether to use pydantic's strict mode when instantiating the
                new model.
        deepcopy: Whether to create deep copies of the fields.

    Returns:
        The converted model.

    """
    common_fields = (
        source.model_fields.keys() & target_type.model_fields.keys()
    )

    data = {field: getattr(source, field) for field in common_fields}

    if source.model_extra is not None:
        data |= source.model_extra

    target = target_type.model_validate(data, strict=strict)

    return copy.deepcopy(target) if deepcopy else target


def _parse_list(
    text: str, /, collection: type[_ListOrTupleT] = list
) -> _ListOrTupleT:
    """Parse a string into a list of strings.

    Items are separated by whitespace or commas.

    Args:
        text: The string to parse.
        collection: The type of collection to return.

    Returns:
        A collection of strings.

    Examples:
        >>> _parse_list("a b c")
        ['a', 'b', 'c']
        >>> _parse_list("a, b, c")
        ['a', 'b', 'c']
        >>> _parse_list("a b, c")
        ['a', 'b', 'c']
        >>> _parse_list("a,b,c")
        ['a', 'b', 'c']
        >>> _parse_list("")
        []

    """
    result = (part for part in re.split(r"\s+|\s*,\s*", text) if part)
    return collection(result)


def _get_validator(
    func: Callable[[str], object], /
) -> pydantic.BeforeValidator:
    def validator(value: object) -> object:
        return func(value) if isinstance(value, str) else value

    return pydantic.BeforeValidator(validator)


List: TypeAlias = Annotated[
    list[_T],
    _get_validator(functools.partial(_parse_list, collection=list)),
]
"""Pydantic field for a list of strings. Uses `_parse_list` as a validator."""

Tuple: TypeAlias = Annotated[
    _T, _get_validator(functools.partial(_parse_list, collection=tuple))
]
"""Pydantic field for a tuple of strings. Uses `_parse_list` as a validator."""

# unfortunately, there doesn't seem to be a better way to do this
# see https://github.com/python/typing/issues/779
Tuple1: TypeAlias = Tuple[tuple[_T1]]
Tuple2: TypeAlias = Tuple[tuple[_T1, _T2]]
Tuple3: TypeAlias = Tuple[tuple[_T1, _T2, _T3]]
Tuple4: TypeAlias = Tuple[tuple[_T1, _T2, _T3, _T4]]


class BaseModel(pydantic.BaseModel):
    """Pydantic BaseModel with sane defaults and a few extra tweaks."""

    model_config = pydantic.ConfigDict(
        allow_inf_nan=False,
        extra="forbid",
        strict=True,
        validate_assignment=True,
        validate_default=True,
        validate_return=True,
    )

    # patch pydantic's __repr__ so that is doesn't break on cyclic models
    # see https://github.com/pydantic/pydantic/issues/9424
    @reprlib.recursive_repr()
    @override
    def __repr__(self) -> str:
        return super().__repr__()

    @override
    def __str__(self) -> str:
        return repr(self)



"""Abstract base classes for representing common XML entities.

This module defines the following classes:
- `Entity`: The base class for all XML entities.
- `Element`: A subclass of `Entity` that represents an XML element.
- `CharacterData`: A subclass of `Entity` that represents a textual entity.
- `CData`, `Comment`, `RawText`: Subclasses of `CharacterData` that represent
    specific types of character data in XML.
"""

from __future__ import annotations

import abc
import collections
import contextlib
import reprlib
import sys
import warnings
from collections.abc import Generator, Mapping

import bs4
import pydantic
from typing_extensions import (
    Final,
    Literal,
    Self,
    SupportsIndex,
    TypeVar,
    cast,
    final,
    overload,
    override,
)

from svglab import constants, errors, models, serialize
from svglab.attrparse import iri, length, transform
from svglab.attrs import attrdefs, attrgroups
from svglab.attrs import names as attr_names
from svglab.elements import names
from svglab.utils import bsutils, iterutils, mathutils, miscutils


_T = TypeVar("_T")
_T_element = TypeVar("_T_element", bound="Element")
_TransformT1 = TypeVar("_TransformT1", bound=transform.TransformFunction)
_TransformT2 = TypeVar("_TransformT2", bound=transform.TransformFunction)

_EMPTY_PARAM: Final = object()
"""A sentinel value for an empty parameter."""


class StrokeWidthScaled:
    """The element's `stroke-width` attribute should be scaled."""


def _match_element(
    element: Element, /, *, search: type[Element] | names.ElementName | str
) -> bool:
    """Check if an element matches the given search criteria.

    Args:
    element: The element to check.
    search: The search criteria. Can be an element name or an element class.

    Returns:
    `True` if the element matches the search criteria, `False` otherwise.

    Examples:
    >>> from svglab import Rect
    >>> rect = Rect()
    >>> _match_element(rect, search="rect")
    True
    >>> _match_element(rect, search=Rect)
    True
    >>> _match_element(rect, search="circle")
    False

    """
    if isinstance(search, type):
        return isinstance(element, search)

    return element_name(element) == search


def _scale_attr(attr: _T, /, by: float) -> _T:
    """Scale an attribute by the given factor.

    This is a helper function for scaling attributes. If the attribute is a
    number or a non-percentage length, it is scaled by the given factor.
    Otherwise, the attribute is left unchanged. If the attribute is a list or
    tuple, the function is applied recursively to each element.

    Args:
    attr: The attribute to scale.
    by: The factor by which to scale the attribute.

    Returns:
    The scaled attribute.

    Examples:
    >>> from svglab import Length
    >>> _scale_attr(Length(10), 2)
    Length(value=20.0, unit=None)
    >>> _scale_attr(None, 2) is None
    True
    >>> _scale_attr(Length(10, "%"), 2)
    Length(value=10.0, unit='%')

    """
    match attr:
        case length.Length(_, "%"):
            return attr
        case int() | float() | length.Length():
            return cast(_T, attr * by)
        case list() | tuple():
            return type(attr)(_scale_attr(item, by) for item in attr)
        case _:
            return attr


def _scale_stroke_width(
    element: attrdefs.StrokeWidthAttr, by: float
) -> None:
    if (
        isinstance(element, attrdefs.VectorEffectAttr)
        and element.vector_effect == "non-scaling-stroke"
    ):
        return

    sw_set = element.stroke_width is not None

    if not sw_set:
        if not isinstance(element, StrokeWidthScaled):
            return

        element.stroke_width = length.Length(1)

    element.stroke_width = _scale_attr(element.stroke_width, by)  # type: ignore[reportAttributeAccessIssue]

    # if stroke-width was not set and the scaled value is 1 (default),
    # remove the attribute
    if (
        not sw_set
        and isinstance(element.stroke_width, length.Length)
        and mathutils.is_close(float(element.stroke_width), 1)
    ):
        element.stroke_width = None


def scale_distance_along_a_path_attrs(element: object, by: float) -> None:
    """Scale distance-along-a-path attributes of the element.

    The attributes are:
    - `stroke-dasharray`
    - `stroke-dashoffset`

    Args:
        element: The element to scale.
        by: The factor by which to scale the attributes.

    """
    if isinstance(element, attrdefs.StrokeDasharrayAttr) and isinstance(
        element.stroke_dasharray, list
    ):
        element.stroke_dasharray = [
            _scale_attr(dash, by) for dash in element.stroke_dasharray
        ]
    if isinstance(element, attrdefs.StrokeDashoffsetAttr):
        element.stroke_dashoffset = _scale_attr(
            element.stroke_dashoffset, by
        )


def _scale(element: object, scale: transform.Scale) -> None:  # noqa: PLR0915
    if not mathutils.is_close(scale.sx, scale.sy):
        raise ValueError("Non-uniform scaling is not supported.")

    factor = scale.sx

    if mathutils.is_close(factor, 1):
        return

    if isinstance(element, attrdefs.WidthAttr):
        element.width = _scale_attr(element.width, factor)
    if isinstance(element, attrdefs.HeightAttr):
        element.height = _scale_attr(element.height, factor)
    if isinstance(element, attrdefs.RAttr):
        element.r = _scale_attr(element.r, factor)
    if isinstance(element, attrdefs.X1Attr):
        element.x1 = _scale_attr(element.x1, factor)
    if isinstance(element, attrdefs.Y1Attr):
        element.y1 = _scale_attr(element.y1, factor)
    if isinstance(element, attrdefs.X2Attr):
        element.x2 = _scale_attr(element.x2, factor)
    if isinstance(element, attrdefs.Y2Attr):
        element.y2 = _scale_attr(element.y2, factor)
    if isinstance(element, attrdefs.RxAttr):
        element.rx = _scale_attr(element.rx, factor)
    if isinstance(element, attrdefs.RyAttr):
        element.ry = _scale_attr(element.ry, factor)
    if isinstance(element, attrdefs.CxAttr):
        element.cx = _scale_attr(element.cx, factor)
    if isinstance(element, attrdefs.CyAttr):
        element.cy = _scale_attr(element.cy, factor)
    if isinstance(element, attrdefs.FxAttr):
        element.fx = _scale_attr(element.fx, factor)
    if isinstance(element, attrdefs.FyAttr):
        element.fy = _scale_attr(element.fy, factor)
    if isinstance(element, attrdefs.FontSizeAttr):
        element.font_size = _scale_attr(element.font_size, factor)
    if (
        isinstance(element, attrdefs.PointsAttr)
        and element.points is not None
    ):
        element.points = [scale @ point for point in element.points]
    if isinstance(element, attrdefs.DAttr) and element.d is not None:
        element.d = scale @ element.d

    # these assignments have to be mutually exclusive, because the
    # type checker doesn't know that x being a <number> implies that x is
    # not a <coordinate> and vice versa
    if isinstance(element, attrdefs.XNumberAttr):  # noqa: SIM114
        element.x = _scale_attr(element.x, factor)
    elif isinstance(element, attrdefs.XCoordinateAttr):  # noqa: SIM114
        element.x = _scale_attr(element.x, factor)
    elif isinstance(element, attrdefs.XListOfCoordinatesAttr):
        element.x = _scale_attr(element.x, factor)

    if isinstance(element, attrdefs.YNumberAttr):  # noqa: SIM114
        element.y = _scale_attr(element.y, factor)
    elif isinstance(element, attrdefs.YCoordinateAttr):  # noqa: SIM114
        element.y = _scale_attr(element.y, factor)
    elif isinstance(element, attrdefs.YListOfCoordinatesAttr):
        element.y = _scale_attr(element.y, factor)

    if isinstance(element, attrdefs.StrokeWidthAttr):
        _scale_stroke_width(element, factor)

    # no need to scale distance-along-a-path attributes if a custom path
    # length is provided because those attributes and pathLength are
    # proportional
    if not isinstance(element, attrdefs.PathLengthAttr):
        scale_distance_along_a_path_attrs(element, factor)

    if isinstance(element, attrdefs.OffsetNumberPercentageAttr):
        element.offset = _scale_attr(element.offset, factor)


def _translate_attr(attr: _T, /, by: float) -> _T:
    """Translate an attribute by the given amount.

    This is a helper function for translating attributes. If the attribute is
    a number or a non-percentage length, it is translated by the given amount.
    Otherwise, the attribute is left unchanged. If the attribute is a list or
    tuple, the function is applied recursively to each element.

    Args:
    attr: The attribute to translate.
    by: The amount by which to translate the attribute.

    Returns:
    The translated attribute.

    Examples:
    >>> from svglab import Length
    >>> _translate_attr(Length(10), 5)
    Length(value=15.0, unit=None)
    >>> _translate_attr(None, 5) is None
    True
    >>> _translate_attr(Length(10, "%"), 5)
    Length(value=10.0, unit='%')

    """
    match attr:
        case length.Length(_, "%"):
            return attr
        case int() | float():
            return cast(_T, attr + by)
        case length.Length():
            return attr + length.Length(by)
        case list() | tuple():
            return type(attr)(_translate_attr(item, by) for item in attr)
        case _:
            return attr


# pyright goes nuts if `element` is annotated as `Element`... seems like a bug
# luckily we don't really care
def _translate(element: object, translate: transform.Translate) -> None:
    tx, ty = translate.tx, translate.ty

    if mathutils.is_close(tx, 0) and mathutils.is_close(ty, 0):
        return

    zero = length.Length.zero()

    # these attributes are mandatory for the respective elements
    if isinstance(element, attrdefs.X1Attr):
        element.x1 = _translate_attr(element.x1, tx)
    if isinstance(element, attrdefs.Y1Attr):
        element.y1 = _translate_attr(element.y1, ty)
    if isinstance(element, attrdefs.X2Attr):
        element.x2 = _translate_attr(element.x2, tx)
    if isinstance(element, attrdefs.Y2Attr):
        element.y2 = _translate_attr(element.y2, ty)

    # but these are not, so if they are not present, we initialize them
    # to 0
    if isinstance(element, attrdefs.CxAttr):
        element.cx = _translate_attr(element.cx or zero, tx)
    if isinstance(element, attrdefs.CyAttr):
        element.cy = _translate_attr(element.cy or zero, ty)

    if isinstance(element, attrdefs.XNumberAttr):
        element.x = _translate_attr(element.x or 0, tx)
    elif isinstance(element, attrdefs.XCoordinateAttr):
        element.x = _translate_attr(element.x or zero, tx)
    elif isinstance(element, attrdefs.XListOfCoordinatesAttr):
        element.x = _translate_attr(element.x, tx)

    if isinstance(element, attrdefs.YNumberAttr):
        element.y = _translate_attr(element.y or 0, ty)
    elif isinstance(element, attrdefs.YCoordinateAttr):
        element.y = _translate_attr(element.y or zero, ty)
    elif isinstance(element, attrdefs.YListOfCoordinatesAttr):
        element.y = _translate_attr(element.y, ty)

    if (
        isinstance(element, attrdefs.PointsAttr)
        and element.points is not None
    ):
        element.points = [translate @ point for point in element.points]

    if isinstance(element, attrdefs.DAttr) and element.d is not None:
        element.d = translate @ element.d

    if isinstance(element, attrdefs.OffsetNumberPercentageAttr):
        element.offset = _translate_attr(element.offset, tx)


def swap_transforms(
    a: _TransformT1, b: _TransformT2, /
) -> tuple[_TransformT2, _TransformT1]:
    """Swap transforms, adjusting parameters so that the result is equal.

    Args:
        a: The first transform.
        b: The second transform.

    Returns:
        A 2-tuple (b', a') where b' and a' are the adjusted transforms.

    Raises:
        SvgTransformSwapError: If the transforms cannot be swapped.

    Examples:
        >>> from svglab.attrparse.transform import Scale, Translate, SkewX
        >>> swap_transforms(Translate(10, 20), Scale(2, 3))
        (Scale(sx=2.0, sy=3.0), Translate(tx=5.0, ty=6.666666666666667))
        >>> swap_transforms(Scale(2, 3), SkewX(45))
        (SkewX(angle=33.690067525979785), Scale(sx=2.0, sy=3.0))
        >>> swap_transforms(SkewX(45), Translate(10, 20))
        (Translate(tx=30.0, ty=20.0), SkewX(angle=45.0))

    """
    match a, b:
        # transformations of the same type
        case (transform.Translate(), transform.Translate()) | (
            transform.Scale(),
            transform.Scale(),
        ):
            return b, a

        # translate <-> scale
        case transform.Translate(tx, ty), transform.Scale(sx, sy) as scale:
            return scale, type(a)(tx / sx, ty / sy)

        case transform.Scale(sx, sy) as scale, transform.Translate(tx, ty):
            return type(b)(sx * tx, sy * ty), scale

        # translate <-> rotate
        case transform.Rotate(angle, cx, cy), transform.Translate(tx, ty):
            return type(b)(tx, ty), type(a)(angle, cx - tx, cy - ty)
        case transform.Translate(tx, ty), transform.Rotate(angle, cx, cy):
            return type(b)(angle, cx + tx, cy + ty), type(a)(tx, ty)

        # scale <-> rotate
        case transform.Rotate(angle, cx, cy), transform.Scale(
            sx, sy
        ) as scale:
            return scale, type(a)(angle, cx / sx, cy / sy)
        case transform.Scale(sx, sy) as scale, transform.Rotate(
            angle, cx, cy
        ):
            return type(b)(angle, cx * sx, cy * sy), scale

        # translate <-> skew
        case transform.SkewX(angle) as skew_x, transform.Translate(tx, ty):
            return type(b)(tx + ty * mathutils.tan(angle), ty), skew_x

        case transform.Translate(tx, ty), transform.SkewX(angle) as skew_x:
            return skew_x, type(a)(tx - ty * mathutils.tan(angle), ty)

        case transform.SkewY(angle) as skew_y, transform.Translate(tx, ty):
            return type(b)(tx, ty + tx * mathutils.tan(angle)), skew_y

        case transform.Translate(tx, ty), transform.SkewY(angle) as skew_y:
            return skew_y, type(a)(tx, ty - tx * mathutils.tan(angle))

        # scale <-> skew
        case transform.Scale(sx, sy) as scale, transform.SkewX(angle):
            if mathutils.is_close(sx, sy):
                return b, a

            angle = mathutils.arctan(sx / sy * mathutils.tan(angle))
            return type(b)(angle), scale

        case transform.SkewX(angle), transform.Scale(sx, sy) as scale:
            if mathutils.is_close(sx, sy):
                return b, a

            angle = mathutils.arctan(sy / sx * mathutils.tan(angle))
            return scale, type(a)(angle)

        case transform.Scale(sx, sy) as scale, transform.SkewY(angle):
            if mathutils.is_close(sx, sy):
                return b, a

            angle = mathutils.arctan(sy / sx * mathutils.tan(angle))
            return type(b)(angle), scale

        case transform.SkewY(angle), transform.Scale(sx, sy) as scale:
            if mathutils.is_close(sx, sy):
                return b, a

            angle = mathutils.arctan(sx / sy * mathutils.tan(angle))

            return scale, type(a)(angle)
        case _:
            raise errors.SvgTransformSwapError(a, b)


def _move_transformation_to_end(
    transform: transform.Transform, index: int
) -> None:
    """Move a transformation to the end of the transform list.

    This function moves a transformation from the given index to the end of
    the list, swapping it with each transformation that follows it.
    The transformations in the list are adjusted so that the result is the
    same. The transformation itself may have its parameters adjusted as well.

    Args:
        transform: A list of transformations.
        index: The index of the transformation to move.

    Raises:
        ValueError: If the index is out of range.
        SvgTransformSwapError: If two transformations cannot be swapped.

    Examples:
        >>> from svglab.attrparse.transform import Translate, Scale
        >>> transform = [Translate(10, 20), Scale(2, 3)]
        >>> _move_transformation_to_end(transform, 0)
        >>> transform
        [Scale(sx=2.0, sy=3.0), Translate(tx=5.0, ty=6.666666666666667)]

    """
    if not (0 <= index < len(transform)):
        msg = f"Index {index=} out of range"
        raise ValueError(msg)

    for i in range(index, len(transform) - 1):
        transform[i], transform[i + 1] = swap_transforms(
            transform[i], transform[i + 1]
        )


def element_name(element: Element, /) -> str:
    """Get the SVG element name of the given element or element class.

    Args:
    element: The element or element class.

    Returns:
    The SVG element name.

    Examples:
    >>> from svglab import Rect
    >>> element_name(Rect())
    'rect'

    """
    if isinstance(element, UnknownElement):
        return element.element_name

    return names.ELEMENT_NAME_TO_NORMALIZED.inverse[type(element).__name__]


class Entity(models.BaseModel, metaclass=abc.ABCMeta):
    """The base class of the SVG element hierarchy."""

    parent: Element | None = pydantic.Field(default=None, init=False)

    def to_xml(
        self,
        *,
        pretty: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> str:
        """Convert the element to XML.

        Args:
        pretty: Whether to produce pretty-printed XML.
        indent: The number of spaces to indent each level of the document.
        formatter: The formatter to use for serialization.

        Returns:
        The XML representation of the element.

        Examples:
        >>> from svglab import Rect, Length
        >>> rect = Rect(id="foo", stroke_linecap="round")
        >>> rect.to_xml()
        '<rect id="foo" stroke-linecap="round"/>'

        """
        with formatter or serialize.get_current_formatter():
            soup = self.to_beautifulsoup_object()
            return bsutils.beautifulsoup_to_str(soup, pretty=pretty)

    @abc.abstractmethod
    def to_beautifulsoup_object(self) -> bs4.PageElement:
        """Convert the element to a corresponding `BeautifulSoup` object."""

    @abc.abstractmethod
    def _eq(self, other: Entity, /) -> bool: ...

    @property
    def ancestors(self) -> Generator[Element]:
        """Iterate over the ancestors of the element.

        The ancestors are the element's parent, grandparent, and so on, up to
        the root ancestor. The root ancestor is the first ancestor that has no
        parent.

        The ancestors are returned in the order from the closest ancestor
        to the root ancestor.

        Yields:
            The ancestors of the element.

        """
        curr = self.parent

        while curr is not None:
            yield curr
            curr = curr.parent

    @override
    def __eq__(self, other: object) -> bool:
        return (
            self._eq(other)
            if miscutils.basic_compare(other, self=self)
            else False
        )


class Element(
    Entity,
    attrgroups.CoreAttrs,
    attrgroups.PresentationAttrs,
    metaclass=abc.ABCMeta,
):
    """An element.

    An element is an entity that has a name, a set of attributes,
    and (optionally) one or more elements as children.

    Elements can have standard attributes (for example, `id`, `class`,
    `color`) and non-standard user-defined attributes.

    Example: `<rect id="foo" class="bar" color="red" />`
    ```
    """

    model_config = pydantic.ConfigDict(
        extra="allow",
        alias_generator=pydantic.AliasGenerator(
            validation_alias=lambda name: pydantic.AliasChoices(
                name,
                attr_names.ATTR_NAME_TO_NORMALIZED.inverse.get(name, name),
            ),
            serialization_alias=(
                lambda name: attr_names.ATTR_NAME_TO_NORMALIZED.inverse.get(
                    name, name
                )
            ),
        ),
    )

    prefix: str | None = None

    __children: list[Entity] = pydantic.PrivateAttr(default_factory=list)

    @pydantic.model_validator(mode="after")
    def __validate_extra(self) -> Element:  # pyright: ignore[reportUnusedFunction]
        # model_extra cannot be None because extra is set to "allow"
        assert self.model_extra is not None, "model_extra is None"

        for key, value in self.model_extra.items():
            if not isinstance(value, str):
                msg = (
                    f"Non-standard attribute {key!r} must be of type str,"
                    f" got {type(value)}"
                )
                raise TypeError(msg)

        return self

    def extra_attrs(self) -> Mapping[str, str]:
        """Get the extra attributes of the element.

        Extra attributes are user-defined attributes that are not part of the
        SVG standard.

        Returns:
            A mapping of extra attribute names to their values. The mapping
            should be considered immutable.

        """
        assert self.model_extra is not None, "model_extra is None"
        return self.model_extra

    def standard_attrs(self) -> Mapping[attr_names.AttributeName, object]:
        """Get the standard attributes of the element.

        Standard attributes are attributes that are part of the SVG standard.

        Returns:
            A mapping of standard attribute names to their values. The mapping
            should be considered immutable.

        """
        dump = self.model_dump(by_alias=True, exclude_none=True)

        return {
            attr: getattr(self, attr_names.ATTR_NAME_TO_NORMALIZED[attr])
            for key, _ in dump.items()
            if (attr := cast(attr_names.AttributeName, key))
            in attr_names.ATTR_NAME_TO_NORMALIZED
        }

    def all_attrs(self) -> Mapping[str, object]:
        """Get all attributes of the element.

        This includes both standard and extra attributes.

        Returns:
            A mapping of attribute names to their values. The mapping should
            be considered immutable.

        """
        standard = cast(Mapping[str, object], self.standard_attrs())
        extra = self.extra_attrs()

        return {**standard, **extra}

    @override
    def _eq(self, other: Entity) -> bool:
        return (
            isinstance(other, Element)
            and self.prefix == other.prefix
            and self.all_attrs() == other.all_attrs()
            and self.num_children == other.num_children
            and all(
                c1 == c2
                for c1, c2 in zip(
                    self.children, other.children, strict=True
                )
            )
        )

    @property
    def children(self) -> Generator[Entity]:
        """Iterate over the children of the element.

        The children are returned in the order they were added to the
        element.

        Yields:
            The children of the element.

        """
        yield from self.__children

    @property
    def descendants(self) -> Generator[Entity]:
        """Iterate over the descendants of the element.

        The descendants of the element are the children and their children,
        and so on. The descendants are returned in a breath-first order.

        Yields:
            The descendants of the element.

        """
        queue = collections.deque(self.children)

        while queue:
            child = queue.popleft()
            yield child

            if isinstance(child, Element):
                queue.extend(child.children)

    @property
    def next_siblings(self) -> Generator[Entity]:
        """Iterate over the next siblings of the element.

        The next siblings are the siblings that come after the element in
        the parent's children list.

        Yields:
            The next siblings of the element.

        """
        if self.parent is None:
            return

        should_yield = False

        for sibling in self.parent.children:
            if should_yield:
                yield sibling
            elif sibling is self:
                should_yield = True

    @property
    def prev_siblings(self) -> Generator[Entity]:
        """Iterate over the previous siblings of the element.

        The previous siblings are the siblings that come before the element
        in the parent's children list.

        Yields:
            The previous siblings of the element.

        """
        if self.parent is None:
            return

        for sibling in self.parent.children:
            if sibling is self:
                return

            yield sibling

    @property
    def siblings(self) -> Generator[Entity]:
        """Iterate over the siblings of the element.

        The siblings are the children of the element's parent, excluding
        the element itself.

        If you wish to include the element itself, use `self.parent.children`

        Yields:
            The siblings of the element.

        """
        yield from self.prev_siblings
        yield from self.next_siblings

    def add_child(
        self, child: Entity, /, *, index: int | None = None
    ) -> Self:
        """Add a child to the element.

        Args:
            child: The child to add.
            index: The index at which to add the child. If `None`, the child
                is added at the end of the children list.

        Returns:
            The element itself.

        Raises:
            ValueError: If the child is the same as the element itself or
                if the child already has a parent.

        """
        if child is self:
            raise ValueError("Cannot add an element as a child of itself.")

        if child.parent is not None:
            raise ValueError(
                "Cannot add a child that already has a parent."
            )

        if index is None:
            self.__children.append(child)
        else:
            self.__children.insert(index, child)

        child.parent = self

        return self

    def add_children(self, *children: Entity) -> Self:
        """Add multiple children to the element.

        Args:
            children: The children to add.

        Returns:
            The element itself.

        Raises:
            ValueError: If any child is the same as the element itself or
                if any child already has a parent.

        """
        for child in children:
            self.add_child(child)

        return self

    def remove_child(self, child: Entity, /) -> Entity:
        """Remove a child from the element.

        The child is compared with the children of the element based on
        identity (reference), not equality. This means that the child must be
        the same object as the one that was added to the element.

        Args:
            child: The child to remove.

        Returns:
            The removed child.

        Raises:
            ValueError: If the child is not a child of the element.

        """
        try:
            index = iterutils.search_by_reference(self.__children, child)
        except ValueError as e:
            raise ValueError("Child not found") from e

        return self.pop_child(index)

    def pop_child(self, index: SupportsIndex = -1, /) -> Entity:
        """Remove a child from the element based on its index.

        Args:
            index: The index of the child to remove. If unspecified, the last
                child is removed.

        Returns:
            The removed child.

        Raises:
            IndexError: If the index is out of range.

        """
        child = self.__children.pop(index)
        child.parent = None

        return child

    def clear_children(self) -> None:
        """Remove all children from the element."""
        for child in self.__children:
            child.parent = None

        self.__children.clear()

    def get_child(self, index: SupportsIndex = -1, /) -> Entity:
        """Get a child of the element based on its index.

        Args:
            index: The index of the child to get. If unspecified, the last
                child is returned.

        Returns:
            The child at the specified index.

        Raises:
            IndexError: If the index is out of range.

        """
        return self.__children[index]

    def get_child_index(
        self,
        child: Entity,
        /,
        start: SupportsIndex = 0,
        stop: SupportsIndex = sys.maxsize,
    ) -> int:
        """Get the index of a child in the element's children list.

        The child is compared with the children of the element based on
        identity (reference), not equality. This means that the child must be
        the same object as the one that was added to the element.

        Args:
            child: The child to find.
            start: The starting index for the search. Defaults to 0.
            stop: The ending index for the search. Defaults to the end of the
                list.

        Returns:
            The index of the child in the children list.

        Raises:
            ValueError: If the child is not found in the list.

        """
        return self.__children.index(child, start, stop)

    @override
    def to_beautifulsoup_object(self) -> bs4.Tag:
        element = bs4.Tag(
            name=element_name(self),
            can_be_empty_element=True,
            prefix=self.prefix,
            is_xml=True,
        )

        element.can_be_empty_element = len(self.__children) == 0

        for key, value in self.all_attrs().items():
            element[key] = serialize.serialize_attr(key, value)

        for child in self.children:
            element.append(child.to_beautifulsoup_object())

        if element_name(self) == "svg":
            formatter = serialize.get_current_formatter()

            if formatter.xmlns == "always":
                element["xmlns"] = constants.SVG_XMLNS
            elif formatter.xmlns == "never":
                del element["xmlns"]

        return element

    def __get_main_transform_attribute(
        self,
    ) -> Literal["transform", "gradientTransform", "patternTransform"]:
        match element_name(self):
            case "linearGradient" | "radialGradient":
                return "gradientTransform"
            case "pattern":
                return "patternTransform"
            case _:
                return "transform"

    @property
    def main_transform(self) -> transform.Transform | None:
        """The main transform attribute of the element.

        For most elements, this is the `transform` attribute. However, certain
        elements such as `linearGradient` have theis own specialized
        transform attribute, in which case the `transform` attribute is
        ignored.

        This attribute is an alias to the "main" or "active" transform
        attribute of the element.

        The main transform attributes for the elements are:
        - `gradientTransform` for `linearGradient` and `radialGradient`
        - `patternTransform` for `pattern`
        - `transform` for all other elements
        """
        transform_attr_name = self.__get_main_transform_attribute()

        return cast(
            transform.Transform | None,
            self.__get_attr_or_default(transform_attr_name),
        )

    def get_root(self) -> Element:
        """Get the top-level ancestor of the element.

        The root ancestor is the first ancestor that has no parent. If the
        element has no parent, it is considered the root ancestor.

        Returns:
            The root ancestor of the element.

        """
        return iterutils.take_last(self.ancestors) or self

    def resolve_iri(self, iri_: iri.Iri, /) -> Element:
        """Resolve a local IRI reference to an element in the document.

        This method attempts to resolve an IRI reference to an element in the
        descendants of this element. The IRI reference must be local.

        Args:
            iri_: The IRI reference to resolve. Must be local.

        Returns:
            The element that the IRI reference points to.

        Raises:
            ValueError: If the IRI reference is not local or if the element
                cannot be found in the document.

        """
        if not iri_.is_local:
            raise ValueError(
                "Unable to resolve a non-local IRI reference in this document."
            )

        try:
            return next(
                tag for tag in self.find_all() if tag.id == iri_.fragment
            )
        except StopIteration as e:
            msg = (
                "Unable to find element by IRI "
                f"{iri_.serialize()} in the document."
            )
            raise ValueError(msg) from e

    def references_other_element(self) -> bool:
        """Check if the element contains IRI reference to another element.

        This method checks if the element contains any attributes that
        reference another element in the document. The reference must be
        local and valid, meaning that if the IRI cannot be resolved to an
        element in the document, the method returns `False`. The entire
        document is searched (not just the descendants of this element).

        A warning is emitted if a dangling IRI reference is found.

        Returns:
            `True` if the element contains a local IRI reference to another
            element in the document, `False` otherwise.

        """
        reference_attr_names: list[attr_names.AttributeName] = [
            "xlink:href",
            "href",
            "fill",
            "stroke",
            "mask",
            "clip-path",
        ]

        for attr_name in reference_attr_names:
            if not hasattr(self, attr_name):
                continue

            attr = self.__get_attr_or_default(attr_name)

            if not (isinstance(attr, iri.Iri) and attr.is_local):
                continue

            try:
                self.get_root().resolve_iri(attr)
            except ValueError:
                warnings.warn(
                    f"Dangling IRI reference {attr_name}={attr.serialize()!r}",
                    stacklevel=2,
                )
            else:
                return True

        return False

    @main_transform.setter
    def main_transform(self, value: transform.Transform | None) -> None:
        transform_attr_name = self.__get_main_transform_attribute()

        setattr(self, transform_attr_name, value)

    def __get_attr_or_default(
        self, attr_name: attr_names.AttributeName
    ) -> object:
        attrs = self.standard_attrs()

        if attr_name in attrs:
            return attrs[attr_name]

        match attr_name:
            case "gradientUnits":
                return "objectBoundingBox"
            case "patternUnits":
                return "objectBoundingBox"
            case "patternContentUnits":
                return "userSpaceOnUse"
            case _:
                return None

    def decompose_transform_origin(self) -> None:
        """Decompose the `transform-origin` attribute into `transform`.

        This function replaces the `transform-origin` attribute with a pair of
        `Translate` transformations in the `transform` attribute. The resulting
        `transform` attribute looks like this:

        ```
        self.transform = [
            Translate(tx, ty),
            *self.transform,
            Translate(-tx, -ty),
        ]
        ```

        If the `transform-origin` attribute is not set, this function does
        nothing.

        Raises:
            SvgTransformOriginError: If the value of the `transform-origin`
                attribute cannot be decomposed.

        """
        if self.transform_origin is None:
            return

        if not (
            isinstance(self.transform_origin, tuple)
            and isinstance(self.transform_origin[0], length.Length)
            and isinstance(self.transform_origin[1], length.Length)
        ):
            raise errors.SvgTransformOriginError(self.transform_origin)

        if not self.main_transform:
            self.main_transform = []

        tx = float(self.transform_origin[0])
        ty = float(self.transform_origin[1])

        self.main_transform.insert(0, transform.Translate(tx, ty))
        self.main_transform.append(transform.Translate(-tx, -ty))

        self.transform_origin = None

    def apply_transformation(
        self, transformation: transform.TransformFunction, /
    ) -> None:
        """Apply a transformation to the attributes of the element.

        Args:
        transformation: The transformation to apply.

        Raises:
        ValueError: If the transformation is not supported.
        SvgLengthConversionError: If a length attribute is not convertible
        to user units.

        """
        match transformation:
            case transform.Translate():
                _translate(self, transformation)
            case transform.Scale():
                _scale(self, transformation)
            case _:
                msg = f"Unsupported transformation: {transformation}"
                raise ValueError(msg)

    def __reify_this(self, *, limit: int = sys.maxsize) -> None:
        if limit < 0:
            raise ValueError("Limit must be a positive integer")

        self.decompose_transform_origin()

        if not self.main_transform:
            return

        transform.decompose_matrices(transform=self.main_transform)

        reified = 0
        i = 0

        while reified < limit and i < len(self.main_transform):
            if not isinstance(self.main_transform[i], transform.Reifiable):
                i += 1
                continue

            # move the transformation to the end of the list where it can
            # be directly applied to the elementf
            _move_transformation_to_end(self.main_transform, i)
            transformation = self.main_transform.pop()

            with contextlib.suppress(ValueError):
                self.apply_transformation(transformation)

            for child in self.find_all(recursive=False):
                if element_name(child) == "stop":
                    continue

                if child.main_transform is None:
                    child.main_transform = []

                # decompose transform-origin before we prepend
                child.decompose_transform_origin()
                child.main_transform.insert(0, transformation)

                child.reify(
                    limit=1,
                    recursive=False,
                    remove_transform_list_if_empty=False,
                )

            reified += 1

    def __can_reify(self) -> bool:
        # transformations on <use> elements cannot be reified.
        # reification would only work if all usages of the referenced element
        # via <use> have equal transform attributes (f.e. if the referenced
        # element is only used once), in which case we would continue with
        # reification on the referenced element (probably not worth the hassle)
        # patternTransforms also cannot be reified (no clue why)
        if element_name(self) in ["use", "pattern"]:
            return False

        main_transform_attr_name = self.__get_main_transform_attribute()

        if main_transform_attr_name != "transform" and self.transform:
            warnings.warn(
                (
                    f"Attribute 'transform' on element {type(self)} has "
                    f"no effect. Use {main_transform_attr_name!r} instead."
                ),
                stacklevel=2,
            )

        if (
            main_transform_attr_name == "gradientTransform"
            and self.__get_attr_or_default("gradientUnits")
            == "objectBoundingBox"
        ):
            return False

        # reification of elements that reference other elements (f.e. paint
        # servers or clipping paths) is problematic. there is a lot of stuff
        # that can go wrong here, so we just disable this for now

        # TODO: this is overly general; find out in which cases this is
        # actually needed
        return not self.references_other_element()

    def reify(
        self,
        *,
        limit: int = sys.maxsize,
        recursive: bool = True,
        remove_transform_list_if_empty: bool = True,
    ) -> None:
        """Apply transformations defined by the `transform` attribute.

        This method takes the transformations defined by the `transform`
        attribute and applies them directly to the coordinate, length, and
        other attributes of the element. The transformations are applied in
        the order in which they are defined. The result of this operation
        should be a visually identical element with the `transform` attribute
        reduced or removed (depending on the `limit` parameter).

        Only `Translate` and `Scale` transformations are can be reified. If
        the `transform` attribute contains other transformations, their
        parameters are adjusted so that `Translate` and `Scale` transformations
        can be applied. Unsupported transformations are ignored.

        If all transformations are successfully applied, the `transform`
        attribute is removed from the element.

        All length values in the element must be convertible to user units.

        Args:
            limit: The maximum number of transformations to apply. If the
                `transform` attribute contains more transformations than the
                limit, the remaining transformations are kept in the attribute
                and not applied. The limit must be a positive integer and is
                applied on a per-element basis.
            recursive: If `True`, the method is called recursively on all
                child elements that support reification.
            remove_transform_list_if_empty: If `True`, the `transform`
                attribute is set to `None` if the list is empty after
                reification.

        Raises:
            ValueError: If the limit is not a positive integer.
            SvgUnitConversionError: If a length value cannot be converted to
                user units.
            SvgTransformOriginError: If the value of the `transform-origin`
                attribute is unsupported.
            SvgLengthConversionError: If a length value cannot be
                converted to user units.

        Examples:
            >>> from svglab import Rect, Length, Translate
            >>> rect = Rect(
            ...     x=Length(10),
            ...     y=Length(20),
            ...     width=Length(20),
            ...     height=Length(40),
            ...     transform=[Translate(5, 5)],
            ... )
            >>> rect.reify()
            >>> rect.x
            Length(value=15.0, unit=None)
            >>> rect.y
            Length(value=25.0, unit=None)
            >>> rect.width
            Length(value=20.0, unit=None)
            >>> rect.height
            Length(value=40.0, unit=None)
            >>> rect.transform is None
            True

        """
        if not self.__can_reify():
            return

        self.__reify_this(limit=limit)

        if remove_transform_list_if_empty and not self.transform:
            self.transform = None

        if recursive:
            for child in self.find_all(recursive=False):
                child.reify(
                    limit=limit,
                    recursive=True,
                    remove_transform_list_if_empty=remove_transform_list_if_empty,
                )

    @overload
    def find_all(
        self, /, *, recursive: bool = True
    ) -> Generator[Element]: ...

    @overload
    def find_all(
        self, *elements: type[_T_element], recursive: bool = True
    ) -> Generator[_T_element]: ...

    @overload
    def find_all(
        self,
        *elements: type[Element] | names.ElementName | str,
        recursive: bool = True,
    ) -> Generator[Element]: ...

    def find_all(
        self,
        *elements: type[Element] | names.ElementName | str,
        recursive: bool = True,
    ) -> Generator[Element]:
        """Find all elements that match the given search criteria.

        Args:
        elements: The elements to search for. Can be element names or element
        classes.  If no search criteria are provided, all elements are
        returned.
        recursive: If `False`, only search the direct children of the element,
        otherwise search all descendants.

        Returns:
        An iterator over all elements that match the search criteria.

        Examples:
        >>> from svglab import G, Rect
        >>> g = G().add_children(Rect(), G().add_child(Rect()))
        >>> list(g.find_all("rect"))
        [Rect(), Rect()]
        >>> list(g.find_all(G))
        [G(children=[Rect()])]
        >>> list(g.find_all(Rect, recursive=False))
        [Rect()]
        >>> list(g.find_all(G, "rect"))
        [Rect(), G(children=[Rect()]), Rect()]

        """
        for child in self.descendants if recursive else self.children:
            if isinstance(child, Element) and (
                not elements
                or any(
                    _match_element(child, search=element)
                    for element in elements
                )
            ):
                yield child

    @overload
    def find(
        self, *elements: type[_T_element], recursive: bool = True
    ) -> _T_element: ...

    @overload
    def find(
        self,
        *elements: type[Element] | names.ElementName | str,
        recursive: bool = True,
    ) -> Element: ...

    @overload
    def find(
        self,
        *elements: type[_T_element],
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> _T_element | _T: ...

    @overload
    def find(
        self,
        *elements: type[Element] | names.ElementName | str,
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> Element | _T: ...

    def find(
        self,
        *elements: type[Element] | names.ElementName | str,
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> Element | _T:
        """Find the first element that matches the given search criteria.

        Args:
        elements: The elements to search for. Can be element names or element
        classes.
        recursive: If `False`, only search the direct children of the element,
        otherwise search all descendants.
        default: The default value to return if no element matches the search
        criteria.

        Returns:
        The first element that matches the search criteria.

        Raises:
        SvgElementNotFoundError: If no element matches the search criteria and
        no default value is provided.

        Examples:
        >>> from svglab import G, Rect
        >>> g = G().add_children(
        ...     Rect(id="foo"), G().add_child(Rect(id="bar"))
        ... )
        >>> g.find("rect")
        Rect(id='foo')
        >>> g.find(G)
        G(children=[Rect(id='bar')])
        >>> g.find("circle", default=None) is None
        True

        """
        try:
            return next(self.find_all(*elements, recursive=recursive))
        except StopIteration as e:
            if default is not _EMPTY_PARAM:
                return default

            msg = f"Unable to find element by search criteria: {elements}"
            raise errors.SvgElementNotFoundError(msg) from e

    @property
    def num_children(self) -> int:
        """Get the number of children of the element.

        Returns:
        The number of children of the element.

        """
        return len(self.__children)

    def has_children(self) -> bool:
        """Check if the element has any children.

        Returns:
        `True` if the element has children, `False` otherwise.

        """
        return self.num_children > 0

    def __getitem__(self, key: str) -> str:
        assert self.model_extra is not None, "model_extra is None"
        value: str = self.model_extra[key]
        return value

    def __setitem__(self, key: str, value: str) -> None:
        assert self.model_extra is not None, "model_extra is None"
        self.model_extra[key] = value

    def __delitem__(self, key: str) -> None:
        assert self.model_extra is not None, "model_extra is None"
        del self.model_extra[key]

    @reprlib.recursive_repr()
    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        attrs = dict(self.all_attrs())

        if self.__children:
            attrs["children"] = list(self.children)

        if self.prefix:
            attrs["prefix"] = self.prefix

        if isinstance(self, UnknownElement):
            attrs["element_name"] = self.element_name

        attr_repr = ", ".join(
            f"{key}={value!r}" for key, value in attrs.items()
        )

        return f"{name}({attr_repr})"

    def get_iri(self) -> iri.Iri:
        """Obtain a local IRI reference to this element.

        The reference is constructed based on the element's id. If the element
        has no id, an exception is raised.

        Returns:
            A local IRI reference to this element.

        Raises:
            RuntimeError: If the element has no id.

        """
        if self.id is None:
            msg = "Unable to create local IRI reference to element with no id"
            raise RuntimeError(msg)

        return iri.Iri(fragment=self.id)

    def get_func_iri(self) -> iri.FuncIri:
        """Obtain a local FuncIRI reference to this element.

        The reference is constructed based on the element's id. If the element
        has no id, an exception is raised.

        Returns:
            A local FuncIRI reference to this element.

        Raises:
            RuntimeError: If the element has no id.

        """
        return self.get_iri().to_func_iri()


@final
class UnknownElement(Element):
    """Represents an unknown element that is not part of the SVG specification.

    This element has no standard parsed attributes. All attributes are treated
    as extra attributes.

    If you need to parse custom elements including their attributes, it is
    recommended to subclass the `Element` class instead; this class is only
    intended as a last-resort fallback so we don't lose information about
    elements the library does not know about.

    """

    element_name: str = pydantic.Field(frozen=True, min_length=1)
    """The name of the element."""


class CharacterData(Entity, metaclass=abc.ABCMeta):
    """The base class of text-based elements.

    Text-based elements are elements that are represented by a single string.

    Common examples of text-based elements in XML are:
    - `CDATA` sections (`CData`)
    - comments (`Comment`)
    - text (`Text`)
    - processing instructions (for example, `<?xml version="1.0"?>`)
    - Document Type Definitions (DTDs; for example, `<!DOCTYPE html>`)
    """

    content: str = pydantic.Field(frozen=True, min_length=1)

    @override
    def _eq(self, other: Entity) -> bool:
        return (
            isinstance(other, CharacterData)
            and self.content == other.content
        )

    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        return f"{name}({self.content!r})"


@final
class CData(CharacterData):
    """A `CDATA` section.

    A `CDATA` section is a block of text that is not parsed by the XML parser,
    but is interpreted verbatim.

    `CDATA` sections are used to include text that contains characters
    that would otherwise be interpreted as XML markup.

    Example: `<![CDATA[<g id="foo"></g>]]>`
    """

    def __init__(self, content: str, /) -> None:
        """Initialize a CDATA section.

        Args:
            content: The text content of the CDATA section (excluding the
                `<![CDATA[` and `]]>` markers).

        """
        super().__init__(content=content)

    @override
    def to_beautifulsoup_object(self) -> bs4.CData:
        return bs4.CData(self.content)


@final
class Comment(CharacterData):
    """A comment.

    A comment is a block of text that is not parsed by the XML parser,
    but is ignored.

    Comments are used to include notes and other information that is not
    intended to be displayed to the user.

    Example: `<!-- This is a comment -->`
    """

    def __init__(self, content: str, /) -> None:
        """Initialize a comment.

        Args:
            content: The text content of the comment (excluding the `<!--` and
                `-->` markers).

        """
        super().__init__(content=content)

    @override
    def to_beautifulsoup_object(self) -> bs4.Comment:
        return bs4.Comment(self.content)


@final
class RawText(CharacterData):
    """A text node.

    A text node is a block of text that is parsed by the XML parser.

    Text nodes are used to include text that is intended to be displayed
    to the user.

    Example: `Hello, world!`
    """

    def __init__(self, content: str, /) -> None:
        """Initialize a text node.

        Args:
            content: The text content of the node.

        """
        super().__init__(content=content)

    @override
    def to_beautifulsoup_object(self) -> bs4.NavigableString:
        return bs4.NavigableString(self.content)



"""Utility protocols for type hinting."""

# ruff: noqa: D101, D102

from __future__ import annotations

import pydantic
import pydantic_core
from typing_extensions import (
    Protocol,
    Self,
    SupportsFloat,
    SupportsIndex,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)


_T_contra = TypeVar("_T_contra", contravariant=True)

_AnyStr_contra = TypeVar("_AnyStr_contra", str, bytes, contravariant=True)
_AnyStr_co = TypeVar("_AnyStr_co", str, bytes, covariant=True)


SupportsFloatOrIndex: TypeAlias = SupportsFloat | SupportsIndex


@runtime_checkable
class CustomSerializable(Protocol):
    """Protocol for objects with special serialization behavior.

    When a `Serializable` object is serialized, its `serialize()` method is
    used to obtain the string representation, instead of using `str()`.
    """

    def serialize(self) -> str:
        """Return an SVG-friendly string representation of this object."""
        ...


@runtime_checkable
class PydanticCompatible(Protocol):
    """A protocol for classes that can be used as pydantic models."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: type, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema: ...


@runtime_checkable
class SupportsRead(Protocol[_AnyStr_co]):
    def read(self, size: int | None = None, /) -> _AnyStr_co: ...


@runtime_checkable
class SupportsWrite(Protocol[_AnyStr_contra]):
    def write(self, data: _AnyStr_contra, /) -> int: ...


@runtime_checkable
class SupportsAdd(Protocol[_T_contra]):
    def __add__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsRAdd(Protocol[_T_contra]):
    def __radd__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsSub(Protocol[_T_contra]):
    def __sub__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsRSub(Protocol[_T_contra]):
    def __rsub__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsMul(Protocol[_T_contra]):
    def __mul__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsRMul(Protocol[_T_contra]):
    def __rmul__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsTrueDiv(Protocol[_T_contra]):
    def __truediv__(self, other: _T_contra, /) -> Self: ...


class SupportsRTrueDiv(Protocol[_T_contra]):
    def __rtruediv__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsRMatmul(Protocol[_T_contra]):
    def __rmatmul__(self, other: _T_contra, /) -> Self: ...


@runtime_checkable
class SupportsNeg(Protocol):
    def __neg__(self) -> Self: ...


@runtime_checkable
class PointLike(SupportsNeg, Protocol):
    x: float
    y: float



"""Utilities for working with BeautifulSoup."""

from collections.abc import Iterable

import bs4
from typing_extensions import Any, cast, override

from svglab import serialize
from svglab.elements import names


class _BsFormatter(bs4.formatter.XMLFormatter):
    def __init__(self) -> None:
        formatter = serialize.get_current_formatter()
        super().__init__(
            # replaces special characters with XML entities (e.g. < -> &lt;)
            entity_substitution=bs4.formatter.EntitySubstitution.substitute_xml,
            indent=formatter.indent,
        )

    @override
    def attributes(self, tag: bs4.Tag) -> Iterable[tuple[str, Any]]:
        formatter = serialize.get_current_formatter()
        order = []

        if tag.name in formatter.attribute_order:
            order = formatter.attribute_order[
                cast(names.ElementName, tag.name)
            ]
        elif "*" in formatter.attribute_order:
            order = formatter.attribute_order["*"]

        order_index = {attr: i for i, attr in enumerate(order)}
        order_length = len(order)

        # first sort by the predefined order, then by name
        return sorted(
            tag.attrs.items(),
            key=lambda item: (
                order_index.get(item[0], order_length),
                item[0],
            ),
        )


def _make_soup(element: bs4.PageElement, /) -> bs4.BeautifulSoup:
    soup = bs4.BeautifulSoup()
    soup.append(element)
    return soup


def beautifulsoup_to_str(
    element: bs4.PageElement, /, *, pretty: bool
) -> str:
    """Convert a BeautifulSoup PageElement to a string.

    The current formatter settings are used to format the output.

    Args:
        element: The BeautifulSoup PageElement to convert.
        pretty: If `True`, the output will be pretty-printed.
            If `False`, the output will be a single line.

    Returns:
        The string representation of the PageElement.

    """
    result: str

    match element, pretty:
        case bs4.NavigableString(), _:
            result = _make_soup(element).decode(formatter=_BsFormatter())
        case bs4.Tag(), True:
            soup = _make_soup(element)

            result = soup.prettify(formatter=_BsFormatter())
        case bs4.Tag(), False:
            result = str(element)
        case _:
            msg = f"Unsupported type: {type(element)}"
            raise TypeError(msg)

    return result.strip()



"""Utilities for converting between units.

This module provides a way to convert between different units of measurement.

The main output of this module is the `make_converter` function, which creates
a converter function that can convert between units based on a conversion
table.

"""

import collections
import functools
from collections.abc import Callable, Mapping

from typing_extensions import (
    Final,
    LiteralString,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)

from svglab import errors


_Unit: TypeAlias = LiteralString | None
_UnitT_co = TypeVar("_UnitT_co", bound=_Unit, covariant=True)


@runtime_checkable
class _HasUnit(Protocol[_UnitT_co]):
    value: Final[float]
    unit: Final[_UnitT_co]

    def __init__(self, value: float, unit: _UnitT_co) -> None: ...


_HasUnitT = TypeVar("_HasUnitT", bound=_HasUnit[_Unit])
_ConversionGraph: TypeAlias = Mapping[_UnitT_co, Mapping[_UnitT_co, float]]

ConversionTable: TypeAlias = Mapping[tuple[_UnitT_co, _UnitT_co], float]
"""
A table of conversion rates between units.

The keys are tuples of the form `(source, target)`, where `source` is the
unit to convert from and `target` is the unit to convert to. The values are
the conversion rates.
"""

Converter: TypeAlias = Callable[[_HasUnitT, _UnitT_co], _HasUnitT]
"""A function that converts a value to a different unit."""


def _table_to_graph(
    conversion_table: ConversionTable[_UnitT_co],
) -> _ConversionGraph[_UnitT_co]:
    graph = collections.defaultdict(dict)

    for (source, target), rate in conversion_table.items():
        graph[source][target] = rate
        graph[target][source] = 1 / rate

    return graph


def _get_conversion_rate(
    source: _UnitT_co,
    target: _UnitT_co,
    *,
    graph: _ConversionGraph[_UnitT_co],
) -> float | None:
    if source == target:
        return 1

    queue = collections.deque([(source, 1.0)])
    visited = {source}

    while queue:
        u, u_rate = queue.popleft()

        for v, v_rate in graph[u].items():
            if v in visited:
                continue

            rate = u_rate * v_rate

            if v == target:
                return rate

            queue.append((v, rate))
            visited.add(v)

    return None


def make_converter(
    conversion_table: ConversionTable[_UnitT_co],
) -> Converter[_HasUnitT, _UnitT_co]:
    """Create a converter function using the given conversion table.

    Args:
        conversion_table: A table of conversion rates.

    Returns:
        A converter function that can convert between units. The function
        accepts an object with a `value` and `unit` attribute, and a target
        unit. It returns a new object with the converted value and unit.
        If the conversion is not possible,
        a `SvgUnitConversionError` is raised.

    """
    graph = _table_to_graph(conversion_table)

    @functools.cache
    def get_conversion_rate(
        source: _UnitT_co, target: _UnitT_co
    ) -> float | None:
        return _get_conversion_rate(source, target, graph=graph)

    def convert(obj: _HasUnitT, unit: _UnitT_co) -> _HasUnitT:
        conversion_rate: float | None = get_conversion_rate(obj.unit, unit)

        if conversion_rate is None:
            raise errors.SvgUnitConversionError(
                original_unit=obj.unit, target_unit=unit
            )

        return type(obj)(obj.value * conversion_rate, unit)

    return convert



"""Representation of a point in 2D space.

This module defines the `Point` class, which represents a point in 2D space.

The `Point` class is used to define the `<list-of-points>` type in SVG.

Use the `Point` class to represent points in SVG. Use `PointType` in Pydantic
fields.
"""

from __future__ import annotations

from collections.abc import Iterator

import lark
import pydantic
from typing_extensions import (
    Annotated,
    Self,
    SupportsComplex,
    TypeAlias,
    final,
    override,
)

from svglab import mixins, protocols, serialize
from svglab.attrparse import parse, transform
from svglab.utils import mathutils, miscutils


@pydantic.dataclasses.dataclass(frozen=True)
class _Point(
    SupportsComplex,
    mixins.FloatMulDiv,
    transform.PointAddSubWithTranslateRMatmul,
    protocols.PointLike,
    protocols.CustomSerializable,
):
    x: float
    y: float

    @override
    def serialize(self) -> str:
        x, y = serialize.serialize(
            self.x, self.y, precision_group="coordinate"
        )
        formatter = serialize.get_current_formatter()

        return f"{x}{formatter.point_separator}{y}"

    @classmethod
    def from_complex(cls, value: complex, /) -> Self:
        return cls(value.real, value.imag)

    @classmethod
    def zero(cls) -> Self:
        return cls(0, 0)

    def line_reflect(self, center: Self) -> Self:
        """Reflect this point across a line defined by a center point.

        Given a center point, this method returns a new point such that:
        - this point, the center point, and the new point are collinear, and
        - the distance between this point and the center point is equal to the
          distance between the center point and the new point.

        Args:
            center: The center of the reflection.

        Returns:
            The reflected point.

        Examples:
        >>> Point(1, 1).line_reflect(Point(0, 0))
        Point(x=-1.0, y=-1.0)
        >>> Point(0, 0).line_reflect(Point(10, 10))
        Point(x=20.0, y=20.0)

        """
        return center + (center - self)

    def __iter__(self) -> Iterator[float]:
        yield self.x
        yield self.y

    @override
    def __mul__(self, scalar: float) -> Self:
        return type(self)(self.x * scalar, self.y * scalar)

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        as_tuple = tuple(self)
        transformed = other.to_affine() @ as_tuple

        return type(self)(*transformed)

    @override
    def __eq__(self, other: object) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        return mathutils.is_close(self.x, other.x) and mathutils.is_close(
            self.y, other.y
        )

    def __bool__(self) -> bool:
        return self != self.zero()

    def __complex__(self) -> complex:
        return complex(self.x, self.y)


@final
class Point(_Point):
    """A point in a 2D plane.

    Attributes:
        x: The x-coordinate of the point.
        y: The y-coordinate of the point.

    Examples:
        >>> point = Point(1, 2)
        >>> point.x
        1.0
        >>> point.y
        2.0
        >>> point == Point(1, 2)
        True
        >>> point == Point(1, 3)
        False
        >>> point += Point(2, 3)
        >>> point
        Point(x=3.0, y=5.0)
        >>> point -= Point(1, 2)
        >>> point
        Point(x=2.0, y=3.0)
        >>> point *= 2
        >>> point
        Point(x=4.0, y=6.0)
        >>> point /= 2
        >>> point
        Point(x=2.0, y=3.0)
        >>> point + point
        Point(x=4.0, y=6.0)

    """

    @override
    def __init__(
        self,
        x: protocols.SupportsFloatOrIndex,
        y: protocols.SupportsFloatOrIndex,
        /,
    ) -> None:
        super().__init__(float(x), float(y))


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Point]):
    number = float
    point = Point


PointType: TypeAlias = Annotated[
    Point,
    parse.get_validator(grammar="point.lark", transformer=_Transformer()),
]



"""Definitions of the <IRI> and <FuncIRI> types."""

from __future__ import annotations

import functools

import pydantic
import rfc3986
from typing_extensions import Annotated, Self, TypeAlias, final, override

from svglab import protocols


@pydantic.dataclasses.dataclass(frozen=True, kw_only=True)
class Iri(protocols.CustomSerializable):
    """Represents the SVG `<IRI>` type.

    The <IRI> type is a string that represents an Internationalized Resource
    Identifier (IRI). An IRI is a generalization of a URL that allows for
    characters outside the ASCII range.

    """

    scheme: str | None = None
    """The scheme of the IRI, e.g. 'http', 'https', 'ftp', etc."""

    authority: str | None = None
    """The authority of the IRI, e.g. 'www.example.com' or 'example.com'."""

    path: str | None = None
    """The path on the server, e.g. '/path/to/resource'."""

    query: str | None = None
    """The query string of the IRI, e.g. '?key=value&key2=value2'."""

    fragment: str | None = None
    """The fragment identifier of the IRI, e.g. '#section1'."""

    @functools.cached_property
    def is_local(self) -> bool:
        """Whether this is a local IRI reference.

        An IRI is considered local if it has a fragment identifier and no
        scheme, authority, path, or query.

        This is useful for identifying references to elements within the same
        document, for example, when using `url(#id)` in `fill` or `stroke`
        attributes.
        """
        match self:
            case Iri(
                scheme=None,
                authority=None,
                path=None,
                query=None,
                fragment=fragment,
            ) if fragment is not None:
                return True
            case _:
                return False

    @functools.cached_property
    def iri(self) -> str:
        """The full IRI as a string."""
        iri = rfc3986.IRIReference(
            scheme=self.scheme,
            authority=self.authority,
            path=self.path,
            query=self.query,
            fragment=self.fragment,
        )

        return iri.unsplit()

    def to_func_iri(self) -> FuncIri:
        """Convert this Iri to a FuncIri."""
        return FuncIri(
            scheme=self.scheme,
            authority=self.authority,
            path=self.path,
            query=self.query,
            fragment=self.fragment,
        )

    @override
    def serialize(self) -> str:
        return self.iri

    @classmethod
    def from_str(cls, iri: str, /) -> Self:
        """Parse an IRI from a string.

        Args:
            iri: The IRI string to parse. For example 'http://example.com'.

        Returns:
            An instance of the Iri class with the parsed components.

        Raises:
            ValueError: If the IRI is invalid.

        """
        try:
            parsed = rfc3986.iri_reference(iri)

            return cls(
                scheme=parsed.scheme,
                authority=parsed.authority,
                path=parsed.path,
                query=parsed.query,
                fragment=parsed.fragment,
            )

        except Exception as e:
            msg = f"Invalid IRI: {iri}"
            raise ValueError(msg) from e


@final
@pydantic.dataclasses.dataclass(frozen=True)
class FuncIri(Iri):
    """Represents the SVG `<FuncIRI>` type.

    <FuncIRI> represents the functional notation for an IRI. The format is
    `url(<IRI>)`, where `<IRI>` is an IRI.
    """

    @functools.cached_property
    def func_iri(self) -> str:
        """The full functional IRI as a string."""
        return f"url({self.iri})"

    def to_iri(self) -> Iri:
        """Convert this FuncIri to an Iri."""
        return Iri(
            scheme=self.scheme,
            authority=self.authority,
            path=self.path,
            query=self.query,
            fragment=self.fragment,
        )

    @override
    def serialize(self) -> str:
        return self.func_iri

    @classmethod
    @override
    def from_str(cls, iri: str, /) -> Self:
        if not iri.startswith("url(") or not iri.endswith(")"):
            msg = f"Invalid FuncIri: {iri}"
            raise ValueError(msg)

        # Remove the url() part
        iri = iri[4:-1].strip()

        return super().from_str(iri)


IriType: TypeAlias = Annotated[
    Iri,
    pydantic.BeforeValidator(
        lambda v: Iri.from_str(v) if isinstance(v, str) else v
    ),
]
FuncIriType: TypeAlias = Annotated[
    FuncIri,
    pydantic.BeforeValidator(
        lambda v: FuncIri.from_str(v) if isinstance(v, str) else v
    ),
]



"""Useful traits for SVG elements.

Traits are mixins that are used to add common functionality to SVG elements.

For example, we want to be able to normalize the `pathLength` attribute of
all shapes in SVG. Therefore, the `Shape` trait contains the `set_path_length`
method. Elements that are shapes inherit from the `Shape` trait and therefore
automatically have the `set_path_length` method.

This approach allows us to define common functionality in one place and reuse
it on different elements.

Most traits are defined based on the "1.6 Definitions" section of the SVG 1.1
specification.
"""

import abc

from typing_extensions import Protocol

from svglab import entities, graphics, models
from svglab.attrparse import path_data
from svglab.attrs import attrdefs, attrgroups


# common attributes are defined directly on the Element class
class Element(entities.Element):
    """An SVG element."""


class _GraphicalOperations(Element):
    def get_bbox(
        self, *, visible_only: bool = False
    ) -> graphics.BBox | None:
        """Compute the bounding box of this element.

        The bounding box is the smallest rectangle that contains the entire
        element. If the element is not visible, the bounding box is `None`.

        Args:
            visible_only: If `True`, only the visible parts of the element are
                considered when computing the bounding box. If `False`, the
                bounding box includes all parts of the element (even if they
                are transparent).

        Returns:
            The bounding box of the element, or `None` if the element is not
            visible. The bounding box is a tuple of the form `(x_min, y_min,
            x_max, y_max)`.

        """
        return (
            graphics.visible_bbox(self)
            if visible_only
            else graphics.bbox(self)
        )

    def get_mask(
        self,
        *,
        visible_only: bool = False,
        width: float | None = None,
        height: float | None = None,
    ) -> graphics.Mask:
        """Create a mask of this element.

        A mask is a 2D boolean array with `True` values where the element is
        located (or visible) in the rendered SVG and `False` values elsewhere.

        Args:
            element: The element to create a mask for.
            visible_only: If `True`, only the visible parts of the element are
                included in the mask. If `False`, the mask includes all parts
                of the element (even if they are transparent).
            width: The width of the mask. If `None`, the width of the root
                SVG element in the tree is used.
            height: The height of the mask. If `None`, the height of the root
                SVG element in the tree is used.

        Returns:
            A 2D boolean array representing the mask of the element.

        """
        return (
            graphics.visible_mask(self, width=width, height=height)
            if visible_only
            else graphics.mask(self, width=width, height=height)
        )


class GraphicsElement(
    _GraphicalOperations,
    attrgroups.GraphicalEventsAttrs,
    entities.StrokeWidthScaled,
    Element,
):
    """A graphics element.

    From the SVG 1.1 specification:
    > "One of the element types that can cause graphics to be drawn onto the
    target canvas."
    """


class Shape(attrdefs.PathLengthAttr, GraphicsElement):
    """A shape.

    From the SVG 1.1 specification:
    > "A graphics element that is defined by some combination of straight lines
    and curves."
    """

    def set_path_length(self, value: float) -> None:
        """Set a new value for the `pathLength` attribute.

        This method sets a new value for the `pathLength` attribute and scales
        the shape's `stroke-dasharray` and `stroke-dashoffset` attributes
        so that the visual appearance of the shape remains unchanged.

        The shape must have the `pathLength` attribute defined. The `Length`
        values of the scaled attributes must be either in percentages (`%`) or
        convertible to user units. Percentage values are not scaled.

        Args:
            value: The new value for the `pathLength` attribute. Must be
                positive.

        Raises:
            ValueError: If the `value` is not positive.
            RuntimeError: If the current path length is `None`.
            SvgUnitConversionError: If the attribute values cannot be converted
                to user units.

        """
        if value <= 0:
            raise ValueError("Path length must be positive")

        if self.pathLength is None:
            raise RuntimeError("Current pathLength must not be None")

        ratio = value / self.pathLength
        entities.scale_distance_along_a_path_attrs(self, ratio)

        self.pathLength = value


class _PathLike(Protocol):
    d: models.Attr[path_data.PathData]


class BasicShape(Shape, metaclass=abc.ABCMeta):
    """A basic shape.

    From the SVG 1.1 specification:
    > "Standard shapes which are predefined in SVG as a convenience for common
    graphical operations."
    """

    @abc.abstractmethod
    def to_path_data(self) -> path_data.PathData:
        """Convert this basic shape into path data.

        The resulting path data produce the same visual result as the original
        basic shape.

        Returns:
            A `PathData` instance representing the path data.

        """
        ...

    @abc.abstractmethod
    def to_path(self) -> _PathLike:
        """Convert this basic shape into a `Path` element.

        The resulting `Path` element's path data produce the same visual result
        as the original basic shape. The `Path` element will have the same
        attributes as the original basic shape.

        Returns:
            A `Path` element representing the basic shape.

        """
        ...


class AnimationElement(
    attrgroups.AnimationEventsAttrs,
    attrgroups.AnimationTimingAttrs,
    Element,
):
    """An animation element.

    From the SVG 1.1 specification:
    > "An animation element is an element that can be used to animate the
    attribute or property value of another element."
    """


class ContainerElement(
    _GraphicalOperations, attrgroups.GraphicalEventsAttrs, Element
):
    """A container element.

    From the SVG 1.1 specification:
    > "An element which can have graphics elements and other container elements
    as child elements."
    """


class DescriptiveElement(Element):
    """A descriptive element.

    From the SVG 1.1 specification:
    > "An element which provides supplementary descriptive information about
    its parent."
    """


class FilterPrimitiveElement(attrgroups.FilterPrimitivesAttrs, Element):
    """A filter primitive element.

    From the SVG 1.1 specification:
    > "A filter primitive element is one that can be used as a child of
    a `filter` element to specify a node in the filter graph."
    """


class GradientElement(Element):
    """A gradient element.

    From the SVG 1.1 specification:
    > "A gradient element is one that defines a gradient paint server."
    """


class GraphicsReferencingElement(Element):
    """A graphics referencing element.

    From the SVG 1.1 specification:
    > "A graphics element which uses a reference to a different document or
    element as the source of its graphical content."
    """


class LightSourceElement(Element):
    """A light source element.

    From the SVG 1.1 specification:
    > "A light source element is one that can specify light source information
    for an `feDiffuseLighting` or `feSpecularLighting` element."
    """


class StructuralElement(Element):
    """A structural element.

    From the SVG 1.1 specification:
    > "The structural elements are those which define the primary structure of
    an SVG document."
    """


class TextContentElement(GraphicsElement):
    """A text content element.

    From the SVG 1.1 specification:
    > "A text content element is an SVG element that causes a text string to be
    rendered onto the canvas."
    """


class TextContentChildElement(Element):
    """A text content child element.

    From the SVG 1.1 specification:
    > "A text content child element is a text content element that is allowed
    as a descendant of another text content element."
    """


class TextContentBlockElement(Element):
    """A text content block element.

    From the SVG 1.1 specification:
    > "A text content block element is a text content element that serves as a
    standalone element for a unit of text, and which may optionally contain
    certain child text content elements (e.g. `tspan`)."
    """



"""Definitions of all SVG attributes."""

# ruff: noqa: N815, D101

import pydantic
from typing_extensions import Annotated, Literal, TypeAlias

from svglab import models, utiltypes
from svglab.attrs import typedefs


class Attr(models.BaseModel):
    pass


class AccentHeightAttr(Attr):
    accent_height: models.Attr[typedefs.Number] = None


class AccumulateAttr(Attr):
    accumulate: models.Attr[typedefs.None_ | Literal["sum"]] = None


class AdditiveAttr(Attr):
    additive: models.Attr[Literal["replace", "sum"]] = None


class AlignmentBaselineAttr(Attr):
    alignment_baseline: models.Attr[
        typedefs.Auto
        | typedefs.Inherit
        | Literal[
            "baseline",
            "before-edge",
            "text-before-edge",
            "middle",
            "central",
            "after-edge",
            "text-after-edge",
            "ideographic",
            "alphabetic",
            "hanging",
            "mathematical",
        ]
    ] = None


class AlphabeticAttr(Attr):
    alphabetic: models.Attr[typedefs.Number] = None


class AmplitudeAttr(Attr):
    amplitude: models.Attr[typedefs.Number] = None


class ArabicFormAttr(Attr):
    arabic_form: models.Attr[
        Literal["initial", "medial", "terminal", "isolated"]
    ] = None


class AscentAttr(Attr):
    ascent: models.Attr[typedefs.Number] = None


class AttributeNameAttr(Attr):
    attributeName: models.Attr[typedefs.Unparsed] = None


class AttributeTypeAttr(Attr):
    attributeType: models.Attr[Literal["CSS", "XML"] | typedefs.Auto] = (
        None
    )


class AzimuthAttr(Attr):
    azimuth: models.Attr[typedefs.Number] = None


class BaseFrequencyAttr(Attr):
    baseFrequency: models.Attr[typedefs.NumberOptionalNumber] = None


class BaselineShiftAttr(Attr):
    baseline_shift: models.Attr[
        Literal["baseline", "sub", "super"]
        | typedefs.Percentage
        | typedefs.Length
        | typedefs.Inherit
    ] = None


class BaseProfileAttr(Attr):
    baseProfile: models.Attr[typedefs.ProfileName] = None


class BboxAttr(Attr):
    bbox: models.Attr[
        models.Tuple4[
            typedefs.Number,
            typedefs.Number,
            typedefs.Number,
            typedefs.Number,
        ]
    ] = None


class BeginAttr(Attr):
    begin: models.Attr[typedefs.BeginValueList] = None


class BiasAttr(Attr):
    bias: models.Attr[typedefs.Number] = None


class ByAttr(Attr):
    by: models.Attr[typedefs.Unparsed] = None


class CalcModeAttr(Attr):
    calcMode: models.Attr[
        Literal["discrete", "linear", "paced", "spline"]
    ] = None


class CapHeightAttr(Attr):
    cap_height: models.Attr[typedefs.Number] = None


class ClassAttr(Attr):
    class_: models.Attr[typedefs.ListOfStrings] = None


class ClipAttr(Attr):
    clip: models.Attr[
        typedefs.Shape | typedefs.Auto | typedefs.Inherit
    ] = None


class ClipPathAttr(Attr):
    clip_path: models.Attr[
        typedefs.FuncIri | typedefs.None_ | typedefs.Inherit
    ] = None


class ClipPathUnitsAttr(Attr):
    clipPathUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class ClipRuleAttr(Attr):
    clip_rule: models.Attr[
        Literal["nonzero", "evenodd"] | typedefs.Inherit
    ] = None


class ColorAttr(Attr):
    color: models.Attr[typedefs.Color | typedefs.Inherit] = None


class ColorInterpolationAttr(Attr):
    color_interpolation: models.Attr[
        typedefs.Auto | Literal["sRGB", "linearRGB"] | typedefs.Inherit
    ] = None


class ColorInterpolationFiltersAttr(Attr):
    color_interpolation_filters: models.Attr[
        typedefs.Auto | Literal["sRGB", "linearRGB"] | typedefs.Inherit
    ] = None


class ColorProfileAttr(Attr):
    color_profile: models.Attr[
        typedefs.Auto
        | Literal["sRGB"]
        | typedefs.Name
        | typedefs.Iri
        | typedefs.Inherit
    ] = None


class ColorRenderingAttr(Attr):
    color_rendering: models.Attr[
        typedefs.Auto
        | Literal["optimizeSpeed", "optimizeQuality"]
        | typedefs.Inherit
    ] = None


class ContentScriptTypeAttr(Attr):
    contentScriptType: models.Attr[typedefs.ContentType] = None


class ContentStyleTypeAttr(Attr):
    contentStyleType: models.Attr[typedefs.ContentType] = None


class CursorAttr(Attr):
    cursor: models.Attr[
        typedefs.CursorValue
        | typedefs.Inherit
        | models.List[typedefs.CursorValue]
    ] = None


class CxAttr(Attr):
    cx: models.Attr[typedefs.Coordinate] = None


class CyAttr(Attr):
    cy: models.Attr[typedefs.Coordinate] = None


class DAttr(Attr):
    d: models.Attr[typedefs.PathData] = None


class DescentAttr(Attr):
    descent: models.Attr[typedefs.Number] = None


class DiffuseConstantAttr(Attr):
    diffuseConstant: models.Attr[typedefs.Number] = None


class DirectionAttr(Attr):
    direction: models.Attr[Literal["ltr", "rtl"] | typedefs.Inherit] = None


class DisplayAttr(Attr):
    display: models.Attr[
        Literal[
            "block",
            "compact",
            "inline-table",
            "inline",
            "list-item",
            "marker",
            "run-in",
            "table-caption",
            "table-cell",
            "table-column-group",
            "table-column",
            "table-footer-group",
            "table-header-group",
            "table-row-group",
            "table-row",
            "table",
        ]
        | typedefs.None_
        | typedefs.Inherit
    ] = None


class DivisorAttr(Attr):
    divisor: models.Attr[typedefs.Number] = None


class DominantBaselineAttr(Attr):
    dominant_baseline: models.Attr[
        typedefs.Auto
        | Literal[
            "use-script",
            "no-change",
            "reset-size",
            "ideographic",
            "alphabetic",
            "hanging",
            "mathematical",
            "central",
            "middle",
            "text-after-edge",
            "text-before-edge",
        ]
        | typedefs.Inherit
    ] = None


class DurAttr(Attr):
    dur: models.Attr[
        typedefs.ClockValue | Literal["media", "indefinite"]
    ] = None


class DxListOfLengthsAttr(Attr):
    dx: models.Attr[typedefs.ListOfLengths] = None


class DxNumberAttr(Attr):
    dx: models.Attr[typedefs.Number] = None


class DyListOfLengthsAttr(Attr):
    dy: models.Attr[typedefs.ListOfLengths] = None


class DyNumberAttr(Attr):
    dy: models.Attr[typedefs.Number] = None


class EdgeModeAttr(Attr):
    edgeMode: models.Attr[Literal["duplicate", "wrap", "none"]] = None


class ElevationAttr(Attr):
    elevation: models.Attr[typedefs.Number] = None


class EnableBackgroundAttr(Attr):
    enable_background: models.Attr[
        Literal["accumulate"]
        | typedefs.Inherit
        | typedefs.Unparsed  # new [ <x> <y> <width> <height> ]
    ] = None


class EndAttr(Attr):
    end: models.Attr[typedefs.EndValueList] = None


class ExponentAttr(Attr):
    exponent: models.Attr[typedefs.Number] = None


class ExternalResourcesRequiredAttr(Attr):
    externalResourcesRequired: models.Attr[typedefs.Boolean] = None


class FillAttr(Attr):
    fill: models.Attr[typedefs.Paint | Literal["freeze", "remove"]] = None


class FillOpacityAttr(Attr):
    fill_opacity: models.Attr[typedefs.OpacityValue | typedefs.Inherit] = (
        None
    )


class FillRuleAttr(Attr):
    fill_rule: models.Attr[
        Literal["nonzero", "evenodd"] | typedefs.Inherit
    ] = None


class FilterAttr(Attr):
    filter: models.Attr[
        typedefs.FuncIri | typedefs.None_ | typedefs.Inherit
    ] = None


class FilterResAttr(Attr):
    filterRes: models.Attr[typedefs.NumberOptionalNumber] = None


class FilterUnitsAttr(Attr):
    filterUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class FloodColorAttr(Attr):
    flood_color: models.Attr[
        Literal["currentColor"]
        | typedefs.Inherit
        | typedefs.IccColor
        | typedefs.Color
    ] = None


class FloodOpacityAttr(Attr):
    flood_opacity: models.Attr[
        typedefs.OpacityValue | typedefs.Inherit
    ] = None


class FontFamilyAttr(Attr):
    font_family: models.Attr[
        models.List[typedefs.FamilyName | typedefs.GenericFamily]
        | typedefs.Inherit
    ] = None


class FontSizeAttr(Attr):
    font_size: models.Attr[
        typedefs.AbsoluteSize
        | typedefs.RelativeSize
        | typedefs.Length
        | typedefs.Percentage
        | typedefs.Inherit
        | typedefs.All
        | typedefs.ListOfLengths
    ] = None


class FontSizeAdjustAttr(Attr):
    font_size_adjust: models.Attr[
        typedefs.Number | typedefs.None_ | typedefs.Inherit
    ] = None


class FontStretchAttr(Attr):
    font_stretch: models.Attr[
        typedefs.All
        | Literal[
            "condensed ",
            "condensed",
            "expanded",
            "extra-condensed",
            "extra-expanded",
            "narrower",
            "normal",
            "semi-condensed",
            "semi-expanded",
            "ultra-condensed",
            "ultra-expanded",
            "wider",
        ]
        | typedefs.Inherit
    ] = None


class FontStyleAttr(Attr):
    font_style: models.Attr[
        typedefs.All
        | models.List[Literal["normal", "italic", "oblique"]]
        | typedefs.Inherit
    ] = None


class FontVariantAttr(Attr):
    font_variant: models.Attr[
        Literal["normal", "small-caps"] | typedefs.Inherit
    ] = None


# a BeforeValidator must be used on non-string literals; pydantic will not
# coerce literal values
_FontWeightInt: TypeAlias = Annotated[
    Literal[100, 200, 300, 400, 500, 600, 700, 800, 900],
    pydantic.BeforeValidator(int),
]


class FontWeightAttr(Attr):
    font_weight: models.Attr[
        typedefs.All
        | typedefs.Inherit
        | models.List[
            Literal["normal", "bold", "bolder", "lighter"] | _FontWeightInt
        ]
    ] = None


class FormatAttr(Attr):
    format: models.Attr[typedefs.Unparsed] = None


class FrAttr(Attr):
    fr: models.Attr[typedefs.Length] = None


class FromAttr(Attr):
    from_: models.Attr[typedefs.Unparsed] = None


class FxAttr(Attr):
    fx: models.Attr[typedefs.Coordinate] = None


class FyAttr(Attr):
    fy: models.Attr[typedefs.Coordinate] = None


class G1Attr(Attr):
    g1: models.Attr[typedefs.ListOfNames] = None


class G2Attr(Attr):
    g2: models.Attr[typedefs.ListOfNames] = None


class GlyphNameAttr(Attr):
    glyph_name: models.Attr[typedefs.ListOfNames] = None


class GlyphOrientationHorizontalAttr(Attr):
    glyph_orientation_horizontal: models.Attr[
        typedefs.Angle | typedefs.Inherit
    ] = None


class GlyphOrientationVerticalAttr(Attr):
    glyph_orientation_vertical: models.Attr[
        typedefs.Auto | typedefs.Angle | typedefs.Inherit
    ] = None


class GlyphRefAttr(Attr):
    glyphRef: models.Attr[typedefs.Unparsed] = None


class GradientTransformAttr(Attr):
    gradientTransform: models.Attr[typedefs.TransformList] = None


class GradientUnitsAttr(Attr):
    gradientUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class HangingAttr(Attr):
    hanging: models.Attr[typedefs.Number] = None


class HeightAttr(Attr):
    height: models.Attr[typedefs.Length] = None


class HorizAdvXAttr(Attr):
    horiz_adv_x: models.Attr[typedefs.Number] = None


class HorizOriginXAttr(Attr):
    horiz_origin_x: models.Attr[typedefs.Number] = None


class HorizOriginYAttr(Attr):
    horiz_origin_y: models.Attr[typedefs.Number] = None


class HrefAttr(Attr):
    href: models.Attr[typedefs.Iri] = None


class IdAttr(Attr):
    id: models.Attr[typedefs.Name] = None


class IdeographicAttr(Attr):
    ideographic: models.Attr[typedefs.Number] = None


class ImageRenderingAttr(Attr):
    image_rendering: models.Attr[
        typedefs.Auto
        | Literal["optimizeSpeed", "optimizeQuality"]
        | typedefs.Inherit
    ] = None


class In2Attr(Attr):
    in2: models.Attr[
        typedefs.Paint
        | typedefs.FilterPrimitiveReference
        | typedefs.Unparsed
    ] = None


class InAttr(Attr):
    in_: models.Attr[
        typedefs.Paint
        | typedefs.FilterPrimitiveReference
        | typedefs.Unparsed
    ] = None


class InterceptAttr(Attr):
    intercept: models.Attr[typedefs.Number] = None


class K1Attr(Attr):
    k1: models.Attr[typedefs.Number] = None


class K2Attr(Attr):
    k2: models.Attr[typedefs.Number] = None


class K3Attr(Attr):
    k3: models.Attr[typedefs.Number] = None


class K4Attr(Attr):
    k4: models.Attr[typedefs.Number] = None


class KAttr(Attr):
    k: models.Attr[typedefs.Number] = None


class KernelMatrixAttr(Attr):
    kernelMatrix: models.Attr[typedefs.ListOfNumbers] = None


class KernelUnitLengthAttr(Attr):
    kernelUnitLength: models.Attr[typedefs.NumberOptionalNumber] = None


class KerningAttr(Attr):
    kerning: models.Attr[
        typedefs.Auto | typedefs.Length | typedefs.Inherit
    ] = None


class KeyPointsAttr(Attr):
    keyPoints: models.Attr[typedefs.ListOfNumbers] = None


class KeySplinesAttr(Attr):
    keySplines: models.Attr[typedefs.Unparsed] = None


class KeyTimesAttr(Attr):
    keyTimes: models.Attr[typedefs.Unparsed] = None


class LangAttr(Attr):
    lang: models.Attr[typedefs.LanguageTag] = None


class LangGlyphAttr(Attr):
    lang: models.Attr[typedefs.LanguageCodes] = None


class LengthAdjustAttr(Attr):
    lengthAdjust: models.Attr[Literal["spacing", "spacingAndGlyphs"]] = (
        None
    )


class LetterSpacingAttr(Attr):
    letter_spacing: models.Attr[
        Literal["normal"] | typedefs.Length | typedefs.Inherit
    ] = None


class LightingColorAttr(Attr):
    lighting_color: models.Attr[
        Literal["currentColor"]
        | typedefs.Color
        | typedefs.Inherit
        | typedefs.Unparsed
    ] = None


class LimitingConeAngleAttr(Attr):
    limitingConeAngle: models.Attr[typedefs.Number] = None


class LocalAttr(Attr):
    local: models.Attr[typedefs.Unparsed] = None


class MarkerEndAttr(Attr):
    marker_end: models.Attr[
        typedefs.FuncIri | typedefs.None_ | typedefs.Inherit
    ] = None


class MarkerHeightAttr(Attr):
    markerHeight: models.Attr[typedefs.Length] = None


class MarkerMidAttr(Attr):
    marker_mid: models.Attr[
        typedefs.FuncIri | typedefs.None_ | typedefs.Inherit
    ] = None


class MarkerStartAttr(Attr):
    marker_start: models.Attr[
        typedefs.FuncIri | typedefs.None_ | typedefs.Inherit
    ] = None


class MarkerUnitsAttr(Attr):
    markerUnits: models.Attr[Literal["strokeWidth", "userSpaceOnUse"]] = (
        None
    )


class MarkerWidthAttr(Attr):
    markerWidth: models.Attr[typedefs.Length] = None


class MaskAttr(Attr):
    mask: models.Attr[
        typedefs.FuncIri | typedefs.None_ | typedefs.Inherit
    ] = None


class MaskContentUnitsAttr(Attr):
    maskContentUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class MaskUnitsAttr(Attr):
    maskUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class MathematicalAttr(Attr):
    mathematical: models.Attr[typedefs.Number] = None


class MaxAttr(Attr):
    max: models.Attr[Literal["media"] | typedefs.ClockValue] = None


class MediaAttr(Attr):
    media: models.Attr[typedefs.MediaDescriptors] = None


class MethodAttr(Attr):
    method: models.Attr[Literal["align", "stretch"]] = None


class MinAttr(Attr):
    min: models.Attr[Literal["media"] | typedefs.ClockValue] = None


class ModeAttr(Attr):
    mode: models.Attr[
        Literal["normal", "multiply", "screen", "darken", "lighten"]
    ] = None


class NameAnythingAttr(Attr):
    name: models.Attr[typedefs.Anything] = None


class NameNameAttr(Attr):
    name: models.Attr[typedefs.Name] = None


class NumOctavesAttr(Attr):
    numOctaves: models.Attr[typedefs.Integer] = None


class OffsetNumberAttr(Attr):
    offset: models.Attr[typedefs.Number] = None


class OffsetNumberPercentageAttr(Attr):
    offset: models.Attr[typedefs.Number | typedefs.Percentage] = None


class OnAbortAttr(Attr):
    onabort: models.Attr[typedefs.Anything] = None


class OnActivateAttr(Attr):
    onactivate: models.Attr[typedefs.Anything] = None


class OnBeginAttr(Attr):
    onbegin: models.Attr[typedefs.Anything] = None


class OnClickAttr(Attr):
    onclick: models.Attr[typedefs.Anything] = None


class OnEndAttr(Attr):
    onend: models.Attr[typedefs.Anything] = None


class OnErrorAttr(Attr):
    onerror: models.Attr[typedefs.Anything] = None


class OnFocusInAttr(Attr):
    onfocusin: models.Attr[typedefs.Anything] = None


class OnFocusOutAttr(Attr):
    onfocusout: models.Attr[typedefs.Anything] = None


class OnLoadAttr(Attr):
    onload: models.Attr[typedefs.Anything] = None


class OnMouseDownAttr(Attr):
    onmousedown: models.Attr[typedefs.Anything] = None


class OnMouseMoveAttr(Attr):
    onmousemove: models.Attr[typedefs.Anything] = None


class OnMouseOutAttr(Attr):
    onmouseout: models.Attr[typedefs.Anything] = None


class OnMouseOverAttr(Attr):
    onmouseover: models.Attr[typedefs.Anything] = None


class OnMouseUpAttr(Attr):
    onmouseup: models.Attr[typedefs.Anything] = None


class OnRepeatAttr(Attr):
    onrepeat: models.Attr[typedefs.Anything] = None


class OnResizeAttr(Attr):
    onresize: models.Attr[typedefs.Anything] = None


class OnScrollAttr(Attr):
    onscroll: models.Attr[typedefs.Anything] = None


class OnUnloadAttr(Attr):
    onunload: models.Attr[typedefs.Anything] = None


class OnZoomAttr(Attr):
    onzoom: models.Attr[typedefs.Anything] = None


class OpacityAttr(Attr):
    opacity: models.Attr[typedefs.OpacityValue | typedefs.Inherit] = None


class OperatorFeCompositeAttr(Attr):
    operator: models.Attr[
        Literal["over", "in", "out", "atop", "xor", "arithmetic"]
    ] = None


class OperatorFeMorphologyAttr(Attr):
    operator: models.Attr[Literal["erode", "dilate"]] = None


class OrderAttr(Attr):
    order: models.Attr[typedefs.NumberOptionalNumber] = None


class OrientAttr(Attr):
    orient: models.Attr[
        typedefs.Auto | Literal["auto-start-reverse"] | typedefs.Angle
    ] = None


class OrientationAttr(Attr):
    orientation: models.Attr[Literal["h", "v"]] = None


class OriginAttr(Attr):
    origin: models.Attr[Literal["default"]] = None


class OverflowAttr(Attr):
    overflow: models.Attr[
        Literal["visible", "hidden", "scroll"]
        | typedefs.Auto
        | typedefs.Inherit
    ] = None


class OverlinePositionAttr(Attr):
    overline_position: models.Attr[typedefs.Number] = None


class OverlineThicknessAttr(Attr):
    overline_thickness: models.Attr[typedefs.Number] = None


class PaintOrderAttr(Attr):
    paint_order: (
        models.Attr[Literal["normal", "fill", "stroke", "markers"]]
        | typedefs.Inherit
    ) = None


class Panose1Attr(Attr):
    panose1: models.Attr[
        models.Tuple[
            tuple[
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
                typedefs.Integer,
            ]
        ]
    ] = None


class PathAttr(Attr):
    path: models.Attr[typedefs.PathData] = None


class PathLengthAttr(Attr):
    pathLength: models.Attr[typedefs.Number] = None


class PatternContentUnitsAttr(Attr):
    patternContentUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class PatternTransformAttr(Attr):
    patternTransform: models.Attr[typedefs.TransformList] = None


class PatternUnitsAttr(Attr):
    patternUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class PointerEventsAttr(Attr):
    pointer_events: models.Attr[
        Literal[
            "visiblePainted",
            "visibleFill",
            "visibleStroke",
            "visible",
            "painted",
            "fill",
            "stroke",
        ]
        | typedefs.All
        | typedefs.None_
        | typedefs.Inherit
    ] = None


class PointsAttr(Attr):
    points: models.Attr[typedefs.ListOfPoints] = None


class PointsAtXAttr(Attr):
    pointsAtX: models.Attr[typedefs.Number] = None


class PointsAtYAttr(Attr):
    pointsAtY: models.Attr[typedefs.Number] = None


class PointsAtZAttr(Attr):
    pointsAtZ: models.Attr[typedefs.Number] = None


class PreserveAlphaAttr(Attr):
    preserveAlpha: models.Attr[typedefs.Boolean] = None


class PreserveAspectRatioAttr(Attr):
    preserveAspectRatio: models.Attr[
        None
        | Literal[
            "xMinYMin",
            "xMidYMin",
            "xMaxYMin",
            "xMinYMid",
            "xMidYMid",
            "xMaxYMid",
            "xMinYMax",
            "xMidYMax",
            "xMaxYMax",
        ]
        | typedefs.Unparsed
    ] = None


class PrimitiveUnitsAttr(Attr):
    primitiveUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class RAttr(Attr):
    r: models.Attr[typedefs.Length] = None


class RadiusAttr(Attr):
    radius: models.Attr[typedefs.NumberOptionalNumber] = None


class RefXAttr(Attr):
    ref_x: models.Attr[
        Literal["left", "center", "right"] | typedefs.Coordinate
    ] = None


class RefYAttr(Attr):
    ref_y: models.Attr[
        Literal["top", "center", "bottom"] | typedefs.Coordinate
    ] = None


class RenderingIntentAttr(Attr):
    rendering_intent: models.Attr[
        Literal[
            "auto",
            "perceptual",
            "relative-colorimetric",
            "saturation",
            "absolute-colorimetric",
        ]
    ] = None


class RepeatCountAttr(Attr):
    repeatCount: models.Attr[
        Literal["indefinite"] | typedefs.NumericValue
    ] = None


class RepeatDurAttr(Attr):
    repeatDur: models.Attr[Literal["indefinite"] | typedefs.ClockValue] = (
        None
    )


class RequiredExtensionsAttr(Attr):
    requiredExtensions: models.Attr[typedefs.ListOfExtensions] = None


class RequiredFeaturesAttr(Attr):
    requiredFeatures: models.Attr[typedefs.ListOfFeatures] = None


class RestartAttr(Attr):
    restart: models.Attr[Literal["always", "whenNotActive", "never"]] = (
        None
    )


class ResultAttr(Attr):
    result: models.Attr[typedefs.FilterPrimitiveReference] = None


class RotateListOfNumbersAttr(Attr):
    rotate: models.Attr[typedefs.ListOfNumbers] = None


class RotateNumberAutoAutoReverseAttr(Attr):
    rotate: (
        models.Attr[typedefs.Number]
        | typedefs.Auto
        | Literal["auto-reverse"]
    ) = None


class RxAttr(Attr):
    rx: models.Attr[typedefs.Length] = None


class RyAttr(Attr):
    ry: models.Attr[typedefs.Length] = None


class ScaleAttr(Attr):
    scale: models.Attr[typedefs.Number] = None


class SeedAttr(Attr):
    seed: models.Attr[typedefs.Number] = None


class ShapeRenderingAttr(Attr):
    shape_rendering: models.Attr[
        typedefs.Auto
        | Literal["optimizeSpeed", "crispEdges", "geometricPrecision"]
        | typedefs.Inherit
    ] = None


class SlopeAttr(Attr):
    slope: models.Attr[typedefs.Number] = None


class SpacingAttr(Attr):
    spacing: models.Attr[Literal["auto", "exact"]] = None


class SpecularConstantAttr(Attr):
    specularConstant: models.Attr[typedefs.Number] = None


class SpecularExponentAttr(Attr):
    specularExponent: models.Attr[typedefs.Number] = None


class SpreadMethodAttr(Attr):
    spreadMethod: models.Attr[Literal["pad", "reflect", "repeat"]] = None


class StartOffsetAttr(Attr):
    startOffset: models.Attr[typedefs.Length] = None


class StdDeviationAttr(Attr):
    stdDeviation: models.Attr[typedefs.NumberOptionalNumber] = None


class StemhAttr(Attr):
    stemh: models.Attr[typedefs.Number] = None


class StemvAttr(Attr):
    stemv: models.Attr[typedefs.Number] = None


class StitchTilesAttr(Attr):
    stitchTiles: models.Attr[Literal["stitch", "noStitch"]] = None


class StopColorAttr(Attr):
    stop_color: models.Attr[
        Literal["currentColor"]
        | typedefs.Inherit
        | typedefs.IccColor
        | typedefs.Color
    ] = None


class StopOpacityAttr(Attr):
    stop_opacity: models.Attr[typedefs.OpacityValue | typedefs.Inherit] = (
        None
    )


class StrikethroughPositionAttr(Attr):
    strikethrough_position: models.Attr[typedefs.Number] = None


class StrikethroughThicknessAttr(Attr):
    strikethrough_thickness: models.Attr[typedefs.Number] = None


class StringAttr(Attr):
    string: models.Attr[typedefs.Anything] = None


class StrokeAttr(Attr):
    stroke: models.Attr[typedefs.Paint] = None


class StrokeDasharrayAttr(Attr):
    stroke_dasharray: models.Attr[
        typedefs.None_ | typedefs.Dasharray | typedefs.Inherit
    ] = None


class StrokeDashoffsetAttr(Attr):
    stroke_dashoffset: models.Attr[
        typedefs.Percentage | typedefs.Length | typedefs.Inherit
    ] = None


class StrokeLinecapAttr(Attr):
    stroke_linecap: models.Attr[
        Literal["butt", "round", "square"] | typedefs.Inherit
    ] = None


class StrokeLinejoinAttr(Attr):
    stroke_linejoin: models.Attr[
        Literal["miter", "round", "bevel", "miter-clip", "arcs"]
        | typedefs.Inherit
    ] = None


class StrokeMiterlimitAttr(Attr):
    stroke_miterlimit: models.Attr[
        typedefs.Miterlimit | typedefs.Inherit
    ] = None


class StrokeOpacityAttr(Attr):
    stroke_opacity: models.Attr[
        typedefs.OpacityValue | typedefs.Inherit
    ] = None


class StrokeWidthAttr(Attr):
    stroke_width: models.Attr[
        typedefs.Percentage | typedefs.Length | typedefs.Inherit
    ] = None


class StyleAttr(Attr):
    style: models.Attr[typedefs.Unparsed] = None


class SurfaceScaleAttr(Attr):
    surfaceScale: models.Attr[typedefs.Number] = None


class SystemLanguageAttr(Attr):
    systemLanguage: models.Attr[typedefs.LanguageCodes] = None


class TableValuesAttr(Attr):
    tableValues: models.Attr[typedefs.ListOfNumbers] = None


class TargetAttr(Attr):
    target: models.Attr[
        Literal["_replace", "_self", "_parent", "_top", "_blank"]
        | typedefs.XmlName
    ] = None


class TargetXAttr(Attr):
    targetX: models.Attr[typedefs.Integer] = None


class TargetYAttr(Attr):
    targetY: models.Attr[typedefs.Integer] = None


class TextAlignAllAttr(Attr):
    text_align_all: models.Attr[
        Literal[
            "start",
            "end",
            "left",
            "right",
            "center",
            "justify",
            "match-parent",
        ]
        | typedefs.Inherit
    ] = None


class TextAlignAttr(Attr):
    text_align: models.Attr[
        Literal[
            "start",
            "end",
            "left",
            "right",
            "center",
            "justify",
            "match-parent",
            "justify-all",
        ]
        | typedefs.Inherit
    ] = None


class TextAlignLastAttr(Attr):
    text_align_last: models.Attr[
        typedefs.Auto
        | Literal[
            "start",
            "end",
            "left",
            "right",
            "center",
            "justify",
            "match-parent",
        ]
        | typedefs.Inherit
    ] = None


class TextAnchorAttr(Attr):
    text_anchor: models.Attr[
        Literal["start", "middle", "end"] | typedefs.Inherit
    ] = None


class TextDecorationAttr(Attr):
    text_decoration: models.Attr[
        typedefs.None_
        | Literal["underline", "overline", "line-through", "blink"]
        | typedefs.Inherit
    ] = None


class TextIndentAttr(Attr):
    text_indent: models.Attr[
        typedefs.Length
        | typedefs.Percentage
        | Literal["each-line", "hanging"]
        | typedefs.Inherit
    ] = None


class TextLengthAttr(Attr):
    textLength: models.Attr[typedefs.Length] = None


class TextRenderingAttr(Attr):
    text_rendering: models.Attr[
        typedefs.Auto
        | Literal[
            "optimizeSpeed", "optimizeLegibility", "geometricPrecision"
        ]
        | typedefs.Inherit
    ] = None


class TitleAttr(Attr):
    title: models.Attr[typedefs.AdvisoryTitle] = None


class ToAttr(Attr):
    to: models.Attr[typedefs.Unparsed] = None


class TransformAttr(Attr):
    transform: models.Attr[typedefs.TransformList] = None


class TransformOriginAttr(Attr):
    transform_origin: models.Attr[typedefs.TransformOrigin] = None


class TypeAnimateTransformAttr(Attr):
    type: models.Attr[
        Literal["translate", "scale", "rotate", "skewX", "skewY"]
    ] = None


class TypeContentTypeAttr(Attr):
    type: models.Attr[typedefs.ContentType] = None


class TypeFeColorMatrixAttr(Attr):
    type: models.Attr[
        Literal["matrix", "saturate", "hueRotate", "luminanceToAlpha"]
    ] = None


class TypeFeFuncAttr(Attr):
    type: models.Attr[
        Literal["identity", "table", "discrete", "linear", "gamma"]
    ] = None


class TypeFeTurbluenceAttr(Attr):
    type: models.Attr[Literal["fractalNoise", "turbulence"]] = None


class U1Attr(Attr):
    u1: models.Attr[models.List[typedefs.Character | typedefs.Urange]] = (
        None
    )


class U2Attr(Attr):
    u2: models.Attr[models.List[typedefs.Character | typedefs.Urange]] = (
        None
    )


class UnderlinePositionAttr(Attr):
    underline_position: models.Attr[typedefs.Number] = None


class UnderlineThicknessAttr(Attr):
    underline_thickness: models.Attr[typedefs.Number] = None


class UnicodeAttr(Attr):
    unicode: models.Attr[typedefs.Anything] = None


class UnicodeBidiAttr(Attr):
    unicode_bidi: models.Attr[
        Literal["normal", "embed", "bidi-override"] | typedefs.Inherit
    ] = None


class UnicodeRangeAttr(Attr):
    unicode_range: models.Attr[models.List[typedefs.Urange]] = None


class UnitsPerEmAttr(Attr):
    units_per_em: models.Attr[typedefs.Number] = None


class VAlphabeticAttr(Attr):
    v_alphabetic: models.Attr[typedefs.Number] = None


class VHangingAttr(Attr):
    v_hanging: models.Attr[typedefs.Number] = None


class VIdeographicAttr(Attr):
    v_ideographic: models.Attr[typedefs.Number] = None


class VMathematicalAttr(Attr):
    v_mathematical: models.Attr[typedefs.Number] = None


class ValuesListAttr(Attr):
    values: models.Attr[typedefs.Unparsed] = None


class ValuesListOfNumbersAttr(Attr):
    values: models.Attr[typedefs.ListOfNumbers] = None


class VectorEffectAttr(Attr):
    vector_effect: models.Attr[
        Literal[
            "non-scaling-stroke",
            "non-scaling-size",
            "non-rotation",
            "fixed-position",
        ]
        | typedefs.None_
    ] = None


class VersionAttr(Attr):
    version: models.Attr[Literal["1.0", "1.1", "1.2"]] = None


class VertAdvYAttr(Attr):
    vert_adv_y: models.Attr[typedefs.Number] = None


class VertOriginXAttr(Attr):
    vert_origin_x: models.Attr[typedefs.Number] = None


class VertOriginYAttr(Attr):
    vert_origin_y: models.Attr[typedefs.Number] = None


class ViewBoxAttr(Attr):
    viewBox: models.Attr[
        models.Tuple4[
            typedefs.Number,
            typedefs.Number,
            typedefs.Number,
            typedefs.Number,
        ]
    ] = None


class ViewTargetAttr(Attr):
    viewTarget: models.Attr[models.List[typedefs.XmlName]] = None


class VisibilityAttr(Attr):
    visibility: models.Attr[
        Literal["visible", "hidden", "collapse"] | typedefs.Inherit
    ] = None


class WhiteSpaceAttr(Attr):
    white_space: models.Attr[
        Literal[
            "normal",
            "pre",
            "nowrap",
            "pre-wrap",
            "break-spaces",
            "pre-line",
        ]
        | typedefs.Inherit
    ] = None


class WidthAttr(Attr):
    width: models.Attr[typedefs.Length] = None


class WidthsAttr(Attr):
    widths: models.Attr[
        models.List[
            typedefs.Urange
            | models.Tuple2[typedefs.Urange, typedefs.Number]
        ]
    ] = None


class WordSpacingAttr(Attr):
    word_spacing: models.Attr[
        Literal["normal"] | typedefs.Length | typedefs.Inherit
    ] = None


class WritingModeAttr(Attr):
    writing_mode: models.Attr[
        Literal["lr-tb", "rl-tb", "tb-rl", "lr", "rl", "tb"]
        | typedefs.Inherit
    ] = None


class X1Attr(Attr):
    x1: models.Attr[typedefs.Coordinate] = None


class X2Attr(Attr):
    x2: models.Attr[typedefs.Coordinate] = None


class XChannelSelectorAttr(Attr):
    xChannelSelector: models.Attr[Literal["R", "G", "B", "A"]] = None


class XCoordinateAttr(Attr):
    x: models.Attr[typedefs.Coordinate] = None


class XHeightAttr(Attr):
    x_height: models.Attr[typedefs.Number] = None


class XListOfCoordinatesAttr(Attr):
    x: models.Attr[typedefs.ListOfCoordinates] = None


class XNumberAttr(Attr):
    x: models.Attr[typedefs.Number] = None


class XlinkActuateOnLoadAttr(Attr):
    xlink_actuate: models.Attr[Literal["onLoad"]] = None


class XlinkActuateOnRequestAttr(Attr):
    xlink_actuate: models.Attr[Literal["onRequest"]] = None


class XlinkArcroleAttr(Attr):
    xlink_arcrole: models.Attr[typedefs.Iri] = None


class XlinkHrefAttr(Attr):
    xlink_href: models.Attr[typedefs.Iri] = None


class XlinkRoleAttr(Attr):
    xlink_role: models.Attr[typedefs.Iri] = None


class XlinkShowAttr(Attr):
    xlink_show: models.Attr[
        Literal["new", "replace", "embed", "other", "none"]
    ] = None


class XlinkTitleAttr(Attr):
    xlink_title: models.Attr[typedefs.Anything] = None


class XlinkTypeAttr(Attr):
    xlink_type: models.Attr[Literal["simple"]] = None


class XmlBaseAttr(Attr):
    xml_base: models.Attr[typedefs.Iri] = None


class XmlLangAttr(Attr):
    xml_lang: models.Attr[typedefs.LanguageId] = None


class XmlSpaceAttr(Attr):
    xml_space: models.Attr[Literal["default", "preserve"]] = None


class XmlnsAttr(Attr):
    xmlns: models.Attr[utiltypes.Xmlns] = None


class Y1Attr(Attr):
    y1: models.Attr[typedefs.Coordinate] = None


class Y2Attr(Attr):
    y2: models.Attr[typedefs.Coordinate] = None


class YChannelSelectorAttr(Attr):
    yChannelSelector: models.Attr[Literal["R", "G", "B", "A"]] = None


class YCoordinateAttr(Attr):
    y: models.Attr[typedefs.Coordinate] = None


class YListOfCoordinatesAttr(Attr):
    y: models.Attr[typedefs.ListOfCoordinates] = None


class YNumberAttr(Attr):
    y: models.Attr[typedefs.Number] = None


class ZAttr(Attr):
    z: models.Attr[typedefs.Number] = None


class ZoomAndPanAttr(Attr):
    zoomAndPan: models.Attr[Literal["disable", "magnify"]] = None
